import traceback

from boofuzz import constants, exception
from .base_monitor import BaseMonitor


class CallbackMonitor(BaseMonitor):
    """
    New-Style Callback monitor that is used in Session to provide callback-arrays.
    It's purpose is to keep the \\*_callbacks arguments in the session class while
    simplifying the implementation of session by forwarding these callbacks to
    the monitor infrastructure.

    The mapping of arguments to method implementations of this class is as follows:

    - restart_callbacks --> target_restart
    - pre_send_callbacks --> pre_send
    - post_test_case_callbacks --> post_send
    - post_start_target_callbacks --> post_start_target

    All other implemented interface members are stubs only, as no corresponding
    arguments exist in session. In any case, it is probably wiser to implement
    a custom Monitor than to use the callback functions.

    .. versionadded:: 0.2.0
    """

    def __init__(self, on_pre_send=None, on_post_send=None, on_restart_target=None, on_post_start_target=None):
        BaseMonitor.__init__(self)

        self.on_pre_send = on_pre_send if on_pre_send is not None else []
        self.on_post_send = on_post_send if on_post_send is not None else []
        self.on_restart_target = on_restart_target if on_restart_target is not None else []
        self.on_post_start_target = on_post_start_target if on_post_start_target is not None else []

    def pre_send(self, target=None, fuzz_data_logger=None, session=None):
        """This method iterates over all supplied pre send callbacks and executes them.
        Their return values are discarded, exceptions are catched and logged, but otherwise
        discarded.
        """
        try:
            for f in self.on_pre_send:
                fuzz_data_logger.open_test_step('Pre_Send callback: "{0}"'.format(f.__name__))
                f(target=target, fuzz_data_logger=fuzz_data_logger, session=session, sock=target)
        except Exception:
            fuzz_data_logger.log_error(
                constants.ERR_CALLBACK_FUNC.format(func_name="pre_send") + traceback.format_exc()
            )

    def post_send(self, target=None, fuzz_data_logger=None, session=None):
        """This method iterates over all supplied post send callbacks and executes them.
        Their return values are discarded, exceptions are caught and logged:

        - :class:`BoofuzzTargetConnectionReset <boofuzz.exception.BoofuzzTargetConnectionReset>` will log a failure
        - :class:`BoofuzzTargetConnectionAborted <boofuzz.exception.BoofuzzTargetConnectionAborted>` will log an info
        - :class:`BoofuzzTargetConnectionFailedError <boofuzz.exception.BoofuzzTargetConnectionFailedError>` will log a
          failure
        - :class:`BoofuzzSSLError <boofuzz.exception.BoofuzzSSLError>` will log either info or failure, depending on
          if the session ignores SSL/TLS errors.
        - every other exception is logged as an error.

        All exceptions are discarded after handling.
        """
        try:
            for f in self.on_post_send:
                fuzz_data_logger.open_test_step('Post-test case callback: "{0}"'.format(f.__name__))
                f(target=target, fuzz_data_logger=fuzz_data_logger, session=session, sock=target)
        except exception.BoofuzzTargetConnectionReset:
            fuzz_data_logger.log_fail(constants.ERR_CONN_RESET_FAIL)
        except exception.BoofuzzTargetConnectionAborted as e:
            fuzz_data_logger.log_info(
                constants.ERR_CONN_ABORTED.format(socket_errno=e.socket_errno, socket_errmsg=e.socket_errmsg)
            )
        except exception.BoofuzzTargetConnectionFailedError:
            fuzz_data_logger.log_fail(constants.ERR_CONN_FAILED)
        except exception.BoofuzzSSLError as e:
            if session._ignore_connection_ssl_errors:
                fuzz_data_logger.log_info(str(e))
            else:
                fuzz_data_logger.log_fail(str(e))
        except Exception:
            fuzz_data_logger.log_error(
                constants.ERR_CALLBACK_FUNC.format(func_name="post_send") + traceback.format_exc()
            )
        finally:
            fuzz_data_logger.open_test_step("Cleaning up connections from callbacks")
        return True

    def restart_target(self, target=None, fuzz_data_logger=None, session=None):
        """
        This Method tries to restart a target. If no restart callbacks are set,
        it returns false; otherwise it returns true.

        :returns: bool
        """
        try:
            for f in self.on_restart_target:
                fuzz_data_logger.open_test_step('Target restart callback: "{0}"'.format(f.__name__))
                f(target=target, fuzz_data_logger=fuzz_data_logger, session=session, sock=target)
        except exception.BoofuzzRestartFailedError:
            raise
        except Exception:
            fuzz_data_logger.log_error(
                constants.ERR_CALLBACK_FUNC.format(func_name="restart_target") + traceback.format_exc()
            )
        finally:
            fuzz_data_logger.open_test_step("Cleaning up connections from callbacks")
            target.close()
            if session._reuse_target_connection:
                fuzz_data_logger.open_test_step("Reopening target connection")
                target.open()

        if len(self.on_restart_target) > 0:
            return True
        else:
            return False

    def post_start_target(self, target=None, fuzz_data_logger=None, session=None):
        """Called after a target is started or restarted."""
        try:
            for f in self.on_post_start_target:
                fuzz_data_logger.open_test_step('Post-start-target callback: "{0}"'.format(f.__name__))
                f(target=target, fuzz_data_logger=fuzz_data_logger, session=session, sock=target)
        except Exception:
            fuzz_data_logger.log_error(
                constants.ERR_CALLBACK_FUNC.format(func_name="post_start_target") + traceback.format_exc()
            )

    def __repr__(self):
        return "CallbackMonitor#{}[pre=[{}],post=[{}],restart=[{}],post_start_target=[{}]]".format(
            id(self),
            ", ".join([x.__name__ for x in self.on_pre_send]),
            ", ".join([x.__name__ for x in self.on_post_send]),
            ", ".join([x.__name__ for x in self.on_restart_target]),
            ", ".join([x.__name__ for x in self.on_post_start_target]),
        )



import itertools
import math
import random

from ..fuzzable import Fuzzable


class String(Fuzzable):
    """Primitive that cycles through a library of "bad" strings.

    The class variable 'fuzz_library' contains a list of
    smart fuzz values global across all instances. The 'this_library' variable contains fuzz values specific to
    the instantiated primitive. This allows us to avoid copying the near ~70MB fuzz_library data structure across
    each instantiated primitive.

    :type name: str, optional
    :param name: Name, for referencing later. Names should always be provided, but if not, a default name will be given,
        defaults to None
    :type default_value: str
    :param default_value: Value used when the element is not being fuzzed - should typically represent a valid value.
    :type size: int, optional
    :param size: Static size of this field, leave None for dynamic, defaults to None
    :type padding: chr, optional
    :param padding: Value to use as padding to fill static field size, defaults to "\\x00"
    :type encoding: str, optional
    :param encoding: String encoding, ex: utf_16_le for Microsoft Unicode, defaults to ascii
    :type max_len: int, optional
    :param max_len: Maximum string length, defaults to None
    :type fuzzable: bool, optional
    :param fuzzable: Enable/disable fuzzing of this primitive, defaults to true
    """

    # store fuzz_library as a class variable to avoid copying the ~70MB structure across each instantiated primitive.
    # Has to be sorted to avoid duplicates
    _fuzz_library = [
        "!@#$%%^#$%#$@#$%$$@#$%^^**(()",
        "",  # strings ripped from spike (and some others I added)
        "$(reboot)",
        "$;reboot",
        "%00",
        "%00/",
        "%01%02%03%04%0a%0d%0aADSF",
        "%01%02%03@%04%0a%0d%0aADSF",
        "%0a reboot %0a",
        "%0Areboot",
        "%0Areboot%0A",
        "%0DCMD=$'reboot';$CMD",
        '%0DCMD=$"reboot";$CMD',
        "%0Dreboot",
        "%0Dreboot%0D",
        "%\xfe\xf0%\x00\xff",
        "%\xfe\xf0%\x01\xff" * 20,
        "%n" * 100,  # format strings.
        "%n" * 500,
        "%s" * 100,
        "%s" * 500,
        "%u0000",
        "& reboot &",
        "& reboot",
        "&&CMD=$'reboot';$CMD",
        '&&CMD=$"reboot";$CMD',
        "&&reboot",
        "&&reboot&&",
        "&CMD=$'reboot';$CMD",
        '&CMD=$"reboot";$CMD',
        "&reboot",
        "&reboot&",
        "'reboot'",
        "..:..:..:..:..:..:..:..:..:..:..:..:..:",
        "/%00/",
        "/." * 5000,
        "/.../" + "B" * 5000 + "\x00\x00",
        "/.../.../.../.../.../.../.../.../.../.../",
        "/../../../../../../../../../../../../boot.ini",
        "/../../../../../../../../../../../../etc/passwd",
        "/.:/" + "A" * 5000 + "\x00\x00",
        "/\\" * 5000,
        "/index.html|reboot|",
        "; reboot",
        ";CMD=$'reboot';$CMD",
        ';CMD=$"reboot";$CMD',
        ";id",
        ";notepad;",
        ";reboot",
        ";reboot/n",
        ";reboot;",
        ";reboot|",
        ";system('reboot')",
        ";touch /tmp/SULLEY;",
        ";|reboot|",
        '<!--#exec cmd="reboot"-->',
        "<>" * 500,  # sendmail crackaddr (http://lsd-pl.net/other/sendmail.txt)
        "<reboot",
        "<reboot%0A",
        "<reboot%0D",
        "<reboot;",
        '"%n"' * 500,
        '"%s"' * 500,
        "\\\\*",
        "\\\\?\\",
        "\nnotepad\n",
        "\nreboot\n",
        "\r\n" * 100,  # miscellaneous.
        "\x01\x02\x03\x04",
        "\xde\xad\xbe\xef" * 10,
        "\xde\xad\xbe\xef" * 100,
        "\xde\xad\xbe\xef" * 1000,
        "\xde\xad\xbe\xef" * 10000,
        "\xde\xad\xbe\xef",  # some binary strings.
        "^CMD=$'reboot';$CMD",
        '^CMD=$"reboot";$CMD',
        "^reboot",
        "`reboot`",
        "a);reboot",
        "a);reboot;",
        "a);reboot|",
        "a)|reboot",
        "a)|reboot;",  # fuzzdb command injection
        "a;reboot",
        "a;reboot;",
        "a;reboot|",
        "a|reboot",
        "CMD=$'reboot';$CMD",
        'CMD=$"reboot";$CMD',
        "FAIL||CMD=$'reboot';$CMD",
        'FAIL||CMD=$"reboot";$CMD',
        "FAIL||reboot",
        "id",
        "id;",
        "id|",
        "reboot",
        "reboot;",
        "reboot|",
        "| reboot",
        "|CMD=$'reboot';$CMD",
        '|CMD=$"reboot";$CMD',
        "|nid",
        "|notepad",
        "|reboot",
        "|reboot;",
        "|reboot|",
        "|touch /tmp/SULLEY",  # command injection.
        "||reboot;",
        "||reboot|",
    ]

    long_string_seeds = [
        "C",
        "1",
        "<",
        ">",
        "'",
        '"',
        "/",
        "\\",
        "?",
        "=",
        "a=",
        "&",
        ".",
        ",",
        "(",
        ")",
        "]",
        "[",
        "%",
        "*",
        "-",
        "+",
        "{",
        "}",
        "\x14",
        "\x00",
        "\xFE",  # expands to 4 characters under utf1
        "\xFF",  # expands to 4 characters under utf1
    ]

    _long_string_lengths = [8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 32768, 0xFFFF]
    _long_string_deltas = [-2, -1, 0, 1, 2]
    _extra_long_string_lengths = [99999, 100000, 500000, 1000000]

    _variable_mutation_multipliers = [2, 10, 100]

    def __init__(
        self, name=None, default_value="", size=None, padding=b"\x00", encoding="utf-8", max_len=None, *args, **kwargs
    ):
        super(String, self).__init__(name=name, default_value=default_value, *args, **kwargs)

        self.size = size
        self.max_len = max_len
        if self.size is not None:
            self.max_len = self.size
        self.encoding = encoding
        self.padding = padding
        if isinstance(padding, str):
            self.padding = self.padding.encode(self.encoding)
        self._static_num_mutations = None
        self.random_indices = {}

        local_random = random.Random(0)  # We want constant random numbers to generate reproducible test cases
        previous_length = 0
        # For every length add a random number of random indices to the random_indices dict. Prevent duplicates by
        # adding only indices in between previous_length and current length.
        for length in self._long_string_lengths:
            self.random_indices[length] = local_random.sample(
                range(previous_length, length), local_random.randint(1, self._long_string_lengths[0])
            )
            previous_length = length

    def _yield_long_strings(self, sequences):
        """
        Given a sequence, yield a number of selectively chosen strings lengths of the given sequence.

        @type  sequences: list(str)
        @param sequences: Sequence to repeat for creation of fuzz strings.
        """
        for sequence in sequences:
            for size in [
                length + delta
                for length, delta in itertools.product(self._long_string_lengths, self._long_string_deltas)
            ]:
                if self.max_len is None or size <= self.max_len:
                    data = sequence * math.ceil(size / len(sequence))
                    yield data[:size]
                else:
                    break

            for size in self._extra_long_string_lengths:
                if self.max_len is None or size <= self.max_len:
                    data = sequence * math.ceil(size / len(sequence))
                    yield data[:size]
                else:
                    break

            if self.max_len is not None:
                data = sequence * math.ceil(self.max_len / len(sequence))
                yield data

        for size in self._long_string_lengths:
            if self.max_len is None or size <= self.max_len:
                s = "D" * size
                for loc in self.random_indices[size]:
                    yield s[:loc] + "\x00" + s[loc + 1 :]  # Replace character at loc with terminator
            else:
                break

    def _yield_variable_mutations(self, default_value):
        for length in self._variable_mutation_multipliers:
            value = default_value * length
            if value not in self._fuzz_library:
                yield value
                if self.max_len is not None and len(value) >= self.max_len:
                    break

    def _adjust_mutation_for_size(self, fuzz_value):
        if self.max_len is not None and self.max_len < len(fuzz_value):
            return fuzz_value[: self.max_len]
        else:
            return fuzz_value

    def mutations(self, default_value):
        """
        Mutate the primitive by stepping through the fuzz library extended with the "this" library, return False on
        completion.

        Args:
            default_value (str): Default value of element.

        Yields:
            str: Mutations
        """
        last_val = None

        for val in itertools.chain(
            self._fuzz_library,
            self._yield_variable_mutations(default_value),
            self._yield_long_strings(self.long_string_seeds),
        ):
            current_val = self._adjust_mutation_for_size(val)
            if last_val == current_val:
                continue
            last_val = current_val
            yield current_val

        # TODO: Add easy and sane string injection from external file/s

    def encode(self, value, mutation_context=None):
        value = value.encode(self.encoding, "replace")
        # pad undersized library items.
        if self.size is not None and len(value) < self.size:
            value += self.padding * (self.size - len(value))
        return value

    def num_mutations(self, default_value):
        """
        Calculate and return the total number of mutations for this individual primitive.

        Args:
            default_value:

        Returns:
            int: Number of mutated forms this primitive can take
        """
        variable_num_mutations = sum(1 for _ in self._yield_variable_mutations(default_value=default_value))
        if self._static_num_mutations is None:
            #  Counting the number of mutations with default value "" results in 0 variable_num_mutations 3 * "" = ""
            self._static_num_mutations = sum(1 for _ in self.mutations(default_value=""))
        return self._static_num_mutations + variable_num_mutations



import errno
import os

from . import itarget_connection


class FileConnection(itarget_connection.ITargetConnection):
    """Writes each message to a new file within the given directory.

    Args:
        directory: Directory for new message files.
        one_file_only (bool): Set to True to continually overwrite a single file. Can be used in conjunction with a hook
            that processes the file.
    """

    def __init__(self, directory, one_file_only=False):
        self._dirname = directory
        self._file_id = 1
        self._file_handle = None
        self._one_file_only = one_file_only

        try:
            os.mkdir(self._dirname)
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise
            pass

    def close(self):
        """
        Close connection to the target.

        :return: None
        """
        self._file_handle.close()
        if not self._one_file_only:
            self._file_id += 1

    def open(self):
        """
        Opens connection to the target. Make sure to call close!

        :return: None
        """
        self._file_handle = open(os.path.join(self._dirname, str(self._file_id)), "wb")

    def recv(self, max_bytes):
        """
        Receive up to max_bytes data from the target.

        Args:
            max_bytes (int): Maximum number of bytes to receive.

        Returns:
            bytes: Received data.
        """
        return b""

    def send(self, data):
        """
        Send data to the target. Only valid after calling open!

        Args:
            data: Data to send.

        Returns:
            int: Number of bytes actually sent.
        """
        self._file_handle.write(data)

    @property
    def info(self):
        return "directory: {0}, filename: {1}".format(self._dirname, str(self._file_id))



from .connection import Connection
from .session import Session, open_test_run
from .session_info import SessionInfo
from .target import Target
from .web_app import WebApp

__all__ = [Connection, SessionInfo, Target, Session, WebApp, open_test_run]



from . import ber, dcerpc, misc, xdr

# all defined legos must be added to this bin.
BIN = {
    "ber_string": ber.String,
    "ber_integer": ber.Integer,
    "dns_hostname": misc.DNSHostname,
    "ndr_conformant_array": dcerpc.NdrConformantArray,
    "ndr_wstring": dcerpc.NdrWString,
    "ndr_string": dcerpc.NdrString,
    "tag": misc.Tag,
    "xdr_string": xdr.String,
}



from . import ifuzz_logger

# IFuzzLoggerBackend is identical to IFuzzLogger at this point, so just copy:
IFuzzLoggerBackend = ifuzz_logger.IFuzzLogger
"""
Interface for backends for FuzzLogger. Currently identical to IFuzzLogger.
"""



class EventHook:
    """
    An EventHook that registers events using +=and -=.

    Based on spassig's solution here: http://stackoverflow.com/a/1094423/461834
    """

    def __init__(self):
        self.__handlers = []

    def __iadd__(self, handler):
        self.__handlers.append(handler)
        return self

    def __isub__(self, handler):
        self.__handlers.remove(handler)
        return self

    def __len__(self):
        return len(self.__handlers)

    def __iter__(self):
        return iter(self.__handlers)

    def fire(self, *args, **kwargs):
        """
        Call each event handler in sequence.

        @param args: Forwarded to event handler.
        @param kwargs: Forwarded to event handler.

        @return: None
        """
        for handler in self.__handlers:
            handler(*args, **kwargs)



import serial

from boofuzz.connections import iserial_like


class SerialConnectionLowLevel(iserial_like.ISerialLike):
    """Internal wrapper for a serial object; backend for SerialConnection.

    Separated from SerialConnection to allow for effective unit testing.

    Implements serial_like.ISerialLike.

    .. versionchanged:: 0.2.0
        SerialConnectionLowLevel has been moved into the connections subpackage.
        The full path is now boofuzz.connections.serial_connection_low_level.SerialConnectionLowLevel
    """

    def __init__(self, port, baudrate, timeout=None):
        """
        @type  port:                   int | str
        @param port:                   Serial port name or number.
        @type baudrate:                int
        @param baudrate:               Baud rate for port.
        @type timeout:                 float
        @param timeout:                Serial port timeout. See pySerial docs. May be updated after creation.
        """
        self._device = None
        self.port = port
        self.baudrate = baudrate
        self.timeout = timeout

    def close(self):
        """
        Close connection to the target.

        :return: None
        """
        self._device.close()

    def open(self):
        """
        Opens connection to the target. Make sure to call close!

        :return: None
        """
        self._device = serial.Serial(port=self.port, baudrate=self.baudrate)

    def recv(self, max_bytes):
        """
        Receive up to max_bytes data from the target.

        :param max_bytes: Maximum number of bytes to receive.
        :type max_bytes: int

        :return: Received data.
        """
        self._device.timeout = self.timeout
        return self._device.read(size=max_bytes)

    def send(self, data):
        """
        Send data to the target. Only valid after calling open!

        :param data: Data to send.

        :return: Number of bytes actually sent.
        """
        return self._device.write(data)



from functools import wraps

from .. import helpers, primitives
from ..fuzzable import Fuzzable


def _may_recurse(f):
    @wraps(f)
    def safe_recurse(self, *args, **kwargs):
        self._recursion_flag = True
        result = f(self, *args, **kwargs)
        self._recursion_flag = False
        return result

    return safe_recurse


class Size(Fuzzable):
    """Create a sizer block bound to the block with the specified name.

    Size blocks that size their own parent or grandparent are allowed.

    :type name: str, optional
    :param name: Name, for referencing later. Names should always be provided, but if not, a default name will be given,
        defaults to None
    :type  block_name:    str, optional
    :param block_name:    Name of block to apply sizer to.
    :type  request:       boofuzz.Request, optional
    :param request:       Request this block belongs to.
    :type  offset:        int, optional
    :param offset:        Offset for calculated size value, defaults to 0
    :type  length:        int, optional
    :param length:        Length of sizer, defaults to 4
    :type  endian:        chr, optional
    :param endian:        Endianness of the bit field (LITTLE_ENDIAN: <, BIG_ENDIAN: >), defaults to LITTLE_ENDIAN
    :type  output_format: str, optional
    :param output_format: Output format, "binary" or "ascii", defaults to binary
    :type  inclusive:     bool, optional
    :param inclusive:     Should the sizer count its own length? Defaults to False
    :type  signed:        bool, optional
    :param signed:        Make size signed vs. unsigned (applicable only with format="ascii"), defaults to False
    :type  math:          def, optional
    :param math:          Apply the mathematical op defined in this function to the size, defaults to None
    :type  fuzzable:      bool, optional
    :param fuzzable:      Enable/disable fuzzing of this block, defaults to true
    """

    def __init__(
        self,
        name=None,
        block_name=None,
        request=None,
        offset=0,
        length=4,
        endian="<",
        output_format="binary",
        inclusive=False,
        signed=False,
        math=None,
        *args,
        **kwargs
    ):
        super(Size, self).__init__(name=name, default_value=None, *args, **kwargs)
        self.block_name = block_name
        self.request = request
        self.offset = offset
        self.length = length
        self.endian = endian
        self.format = output_format
        self.inclusive = inclusive
        self.signed = signed
        self.math = math

        self.bit_field = primitives.BitField(
            name="innerBitField",
            default_value=0,
            width=self.length * 8,
            endian=self.endian,
            output_format=self.format,
            signed=self.signed,
        )
        self._rendered = b""
        self._fuzz_complete = False

        if not self.math:
            self.math = lambda x: x

        # Set the recursion flag before calling a method that may cause a recursive loop.
        self._recursion_flag = False

    def mutations(self, default_value):
        for mutation in self.bit_field.mutations(None):
            yield mutation

    def num_mutations(self, default_value):
        """
        Wrap the num_mutations routine of the internal bit_field primitive.

        :param default_value:
        :rtype:  int
        :return: Number of mutated forms this primitive can take.
        """

        return self.bit_field.get_num_mutations()

    def encode(self, value, mutation_context):
        if value is None:  # default
            if self._recursion_flag:
                return self._get_dummy_value()
            else:
                return helpers.str_to_bytes(
                    self._length_to_bytes(self._calculated_length(mutation_context=mutation_context))
                )
        else:
            return self.bit_field.encode(value=value, mutation_context=mutation_context)

    def _get_dummy_value(self):
        return self.length * b"\x00"

    def _calculated_length(self, mutation_context):
        return (
            self.offset
            + self._inclusive_length_of_self
            + self._length_of_target_block(mutation_context=mutation_context)
        )

    def _length_to_bytes(self, length):
        return primitives.BitField._render_int(
            value=self.math(length),
            output_format=self.format,
            bit_width=self.length * 8,
            endian=self.endian,
            signed=self.signed,
        )

    @property
    def _inclusive_length_of_self(self):
        """Return length of self or zero if inclusive flag is False."""
        if self.inclusive:
            return self.length
        else:
            return 0

    @_may_recurse
    def _length_of_target_block(self, mutation_context):
        """Return length of target block, including mutations if mutation applies."""
        if self.request is not None and self.block_name is not None:
            target_block = self.request.resolve_name(self.context_path, self.block_name)
            return len(target_block.render(mutation_context=mutation_context))
        else:
            return 0

    @property
    @_may_recurse
    def _original_length_of_target_block(self):
        """Return length of target block, including mutations if it is currently mutated."""
        if self.request is not None and self.block_name is not None:
            target_block = self.request.resolve_name(self.context_path, self.block_name)
            length = len(target_block.original_value)
            return length
        else:
            return 0

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self._name)

    def __len__(self):
        return self.length



from ..fuzzable import Fuzzable


class Simple(Fuzzable):
    """Simple bytes value with manually specified fuzz values only.

    :type name: str, optional
    :param name: Name, for referencing later. Names should always be provided, but if not, a default name will be given,
        defaults to None
    :type default_value: Raw, optional
    :param default_value: Raw static data
    :type fuzz_values: list, optional
    :param fuzz_values: List of fuzz values, defaults to None. If empty, Simple is equivalent to Static.
    :type  fuzzable: bool, optional
    :param fuzzable: Enable/disable fuzzing of this primitive, defaults to true
    """

    def __init__(self, name=None, default_value=None, fuzz_values=None, *args, **kwargs):
        super(Simple, self).__init__(name=name, default_value=default_value, fuzz_values=fuzz_values, *args, **kwargs)



from ..fuzzable_block import FuzzableBlock


class Aligned(FuzzableBlock):
    """FuzzableBlock that aligns its contents to a certain number of bytes

    :type  name:        str, optional
    :param name:        Name, for referencing later. Names should always be provided, but if not, a default name will
                        be given, defaults to None
    :type  modulus:     int, optional
    :param modulus:     Pad length of child content to this many bytes, defaults to 1
    :type  request:     boofuzz.Request, optional
    :param request:     Request this block belongs to
    :type  pattern:     bytes, optional
    :param pattern:     Pad using these byte(s)
    :type  fuzzable:    bool, optional
    :param fuzzable:    Enable/disable fuzzing of this block, defaults to true
    """

    def __init__(self, name=None, modulus=1, request=None, pattern=b"\x00", *args, **kwargs):
        super(Aligned, self).__init__(name=name, default_value=None, request=request, *args, **kwargs)
        self._modulus = modulus
        self._pattern = pattern

    def encode(self, value, mutation_context):
        child_data = self.get_child_data(mutation_context=mutation_context)
        padding_length = self._modulus - (len(child_data) % self._modulus)
        a, b = divmod(padding_length, len(self._pattern))
        return child_data + self._pattern * a + self._pattern[:b]



from ..fuzzable_block import FuzzableBlock
from typing import List


class Block(FuzzableBlock):
    """The basic building block. Can contain primitives, sizers, checksums or other blocks.

    :param name: Name, for referencing later. Names should always be provided, but if not, a default name will be given,
        defaults to None
    :type name: str, optional
    :param default_value: Value used when the element is not being fuzzed - should typically represent a valid value,
        defaults to None
    :type default_value: Any, optional
    :param request: Request this block belongs to, defaults to None
    :type request: boofuzz.Request, optional
    :param children: Children of this block, defaults to None
    :type children: boofuzz.Fuzzable, optional
    :param group: Name of group to associate this block with, defaults to None
    :type group: str, optional
    :param encoder: Optional pointer to a function to pass rendered data to prior to return, defaults to None
    :type encoder: callable, optional
    :param dep: Optional primitive whose specific value this block is dependant on, defaults to None
    :type dep: str, optional
    :param dep_value: Value that field "dep" must contain for block to be rendered, defaults to None
    :type dep_value: bytes, optional
    :param dep_values: Values that field "dep" may contain for block to be rendered, defaults to None
    :type dep_values: list, optional
    :param dep_compare: Comparison method to apply to dependency (==, !=, >, >=, <, <=), defaults to None
    :type dep_compare: str, optional
    """

    def __init__(
        self,
        name=None,
        default_value=None,
        request=None,
        children=None,
        group=None,
        encoder=None,
        dep=None,
        dep_value=None,
        dep_values=None,
        dep_compare="==",
        *args,
        **kwargs
    ):
        if dep_value is not None and not isinstance(dep_value, bytes):
            raise TypeError("dep_value must be of bytes type")
        if dep_values is not None and not (
            isinstance(dep_values, list) and all(isinstance(x, bytes) for x in dep_values)
        ):
            raise TypeError("dep_values must be of list of bytes type")

        super(Block, self).__init__(
            name=name, default_value=default_value, request=request, children=children, *args, **kwargs
        )

        self.request = request
        self.group = group
        self.encoder = encoder
        self.dep = dep
        self.dep_value = dep_value
        self.dep_values = dep_values
        self.dep_compare = dep_compare

        self._rendered = b""  # rendered block contents.
        self.group_idx = 0  # if this block is tied to a group, the index within that group.
        self._fuzz_complete = False  # whether or not we are done fuzzing this block.
        self._mutant_index = 0  # current mutation index.

    def mutations(self, default_value, skip_elements=None):
        for item in self.stack:
            self.request.mutant = item
            for mutations in item.get_mutations():
                yield mutations
        if self.group is not None:
            group = self.request.resolve_name(self.context_path, self.group)
            for group_mutations in group.get_mutations():
                for item in self.stack:
                    self.request.mutant = item
                    for mutations in item.get_mutations():
                        yield group_mutations + mutations

    def num_mutations(self, default_value=None):
        n = super(Block, self).num_mutations(default_value=default_value)
        if self.group is not None:
            n += n * self.request.resolve_name(self.context_path, self.group).get_num_mutations()
        return n

    def _do_dependencies_allow_render(self, mutation_context):
        if self.dep:
            dependent_value = self.request.resolve_name(self.context_path, self.dep).get_value(mutation_context)
            if self.dep_compare == "==":
                if self.dep_values and dependent_value not in self.dep_values:
                    return False
                elif not self.dep_values and dependent_value != self.dep_value:
                    return False

            if self.dep_compare == "!=":
                if self.dep_values and dependent_value in self.dep_values:
                    return False
                elif dependent_value == self.dep_value:
                    return False

            if self.dep_compare == ">" and self.dep_value <= dependent_value:
                return False

            if self.dep_compare == ">=" and self.dep_value < dependent_value:
                return False

            if self.dep_compare == "<" and self.dep_value >= dependent_value:
                return False

            if self.dep_compare == "<=" and self.dep_value > dependent_value:
                return False
        return True

    def encode(self, value, mutation_context):
        if self._do_dependencies_allow_render(mutation_context=mutation_context):
            child_data = super(Block, self).get_child_data(mutation_context=mutation_context)
        else:
            child_data = b""
        if self.encoder:
            return self.encoder(child_data)
        else:
            return child_data



import collections
import datetime
import sqlite3
import sys

from . import data_test_case, data_test_step, exception, helpers, ifuzz_logger_backend

# fixup for buffer in python 3
if sys.version_info.major > 2:
    buffer = memoryview


def hex_to_hexstr(input_bytes):
    """
    Render input_bytes as ASCII-encoded hex bytes, followed by a best effort
    utf-8 rendering.

    :param input_bytes: Arbitrary bytes.

    :return: Printable string.
    """
    return helpers.hex_str(input_bytes)


DEFAULT_HEX_TO_STR = hex_to_hexstr


def get_time_stamp():
    s = datetime.datetime.utcnow().isoformat()
    return s


class FuzzLoggerDb(ifuzz_logger_backend.IFuzzLoggerBackend):
    """
    Log fuzz data in a sqlite database file.
    Using an existing database requires more graceful exits to prevent case number duplication.
    """

    def __init__(self, db_filename, num_log_cases=0):
        # Check if a db exists before the connect leaves an "empty" file behind.
        db_already_exist = helpers.path_exists(db_filename)
        self._database_connection = sqlite3.connect(db_filename, check_same_thread=False)
        self._db_cursor = self._database_connection.cursor()

        if not db_already_exist:
            self._db_cursor.execute("""CREATE TABLE cases (name text, number integer, timestamp TEXT)""")
            self._db_cursor.execute(
                """CREATE TABLE steps (test_case_index integer, type text, description text, data blob,
                                    timestamp TEXT, is_truncated BOOLEAN)"""
            )

        self._current_test_case_index = 0

        self._queue = collections.deque([])  # Queue that holds last n test cases before commiting
        self._queue_max_len = num_log_cases
        self._fail_detected = False
        self._log_first_case = True
        self._data_truncate_length = 512

    def get_test_case_data(self, index):
        c = self._database_connection.cursor()
        try:
            test_case_row = next(c.execute("""SELECT * FROM cases WHERE number=?""", [index]))
        except StopIteration:
            return None
        rows = c.execute("""SELECT * FROM steps WHERE test_case_index=?""", [index])
        steps = []
        for row in rows:
            data = row[3]
            # Little hack since BLOB becomes type buffer in py2 and bytes in py3
            # At the end, data will be equivalent types: bytes in py3 and str in py2
            try:
                if isinstance(data, buffer):
                    data = str(data)
            except NameError as e:
                if "buffer" in str(e):  # buffer type does not exist in py3
                    pass
                else:
                    raise
            steps.append(
                data_test_step.DataTestStep(
                    type=row[1], description=row[2], data=data, timestamp=row[4], truncated=row[5]
                )
            )
        return data_test_case.DataTestCase(
            name=test_case_row[0], index=test_case_row[1], timestamp=test_case_row[2], steps=steps
        )

    def open_test_case(self, test_case_id, name, index, *args, **kwargs):
        self._queue.append(["INSERT INTO cases VALUES(?, ?, ?);\n", name, index, helpers.get_time_stamp()])
        self._current_test_case_index = index

    def open_test_step(self, description):
        self._queue.append(
            [
                "INSERT INTO steps VALUES(?, ?, ?, ?, ?, ?);\n",
                self._current_test_case_index,
                "step",
                description,
                b"",
                helpers.get_time_stamp(),
                False,
            ]
        )

    def log_check(self, description):
        self._queue.append(
            [
                "INSERT INTO steps VALUES(?, ?, ?, ?, ?, ?);\n",
                self._current_test_case_index,
                "check",
                description,
                b"",
                helpers.get_time_stamp(),
                False,
            ]
        )

    def log_error(self, description):
        self._queue.append(
            [
                "INSERT INTO steps VALUES(?, ?, ?, ?, ?, ?);\n",
                self._current_test_case_index,
                "error",
                description,
                b"",
                helpers.get_time_stamp(),
                False,
            ]
        )
        self._fail_detected = True
        self._write_log()

    def log_recv(self, data):
        self._queue.append(
            [
                "INSERT INTO steps VALUES(?, ?, ?, ?, ?, ?);\n",
                self._current_test_case_index,
                "receive",
                "",
                buffer(data),
                helpers.get_time_stamp(),
                False,
            ]
        )

    def log_send(self, data):
        self._queue.append(
            [
                "INSERT INTO steps VALUES(?, ?, ?, ?, ?, ?);\n",
                self._current_test_case_index,
                "send",
                "",
                buffer(data),
                helpers.get_time_stamp(),
                False,
            ]
        )

    def log_info(self, description):
        self._queue.append(
            [
                "INSERT INTO steps VALUES(?, ?, ?, ?, ?, ?);\n",
                self._current_test_case_index,
                "info",
                description,
                b"",
                helpers.get_time_stamp(),
                False,
            ]
        )

    def log_fail(self, description=""):
        self._queue.append(
            [
                "INSERT INTO steps VALUES(?, ?, ?, ?, ?, ?);\n",
                self._current_test_case_index,
                "fail",
                description,
                b"",
                helpers.get_time_stamp(),
                False,
            ]
        )
        self._fail_detected = True

    def log_pass(self, description=""):
        self._queue.append(
            [
                "INSERT INTO steps VALUES(?, ?, ?, ?, ?, ?);\n",
                self._current_test_case_index,
                "pass",
                description,
                b"",
                helpers.get_time_stamp(),
                False,
            ]
        )

    def close_test_case(self):
        self._write_log(force=False)

    def close_test(self):
        self._write_log(force=True)

    def _write_log(self, force=False):
        if len(self._queue) > 0:
            if self._queue_max_len > 0:
                while (
                    self._current_test_case_index - next(x for x in self._queue[0] if isinstance(x, int))
                ) >= self._queue_max_len:
                    self._queue.popleft()
            else:
                force = True

            if force or self._fail_detected or self._log_first_case:
                for query in self._queue:
                    # abbreviate long entries first
                    if not self._fail_detected:
                        self._truncate_send_recv(query)
                    self._db_cursor.execute(query[0], query[1:])
                self._queue.clear()
                self._database_connection.commit()
                self._log_first_case = False
                self._fail_detected = False

    def _truncate_send_recv(self, query):
        if query[2] in ["send", "recv"] and len(query[4]) > self._data_truncate_length:
            query[6] = True
            query[4] = buffer(query[4][: self._data_truncate_length])


class FuzzLoggerDbReader:
    """Read fuzz data saved using FuzzLoggerDb

    Args:
        db_filename (str): Name of database file to read.
    """

    def __init__(self, db_filename):
        self._database_connection = sqlite3.connect(db_filename, check_same_thread=False)
        self._db_cursor = self._database_connection.cursor()

    def get_test_case_data(self, index):
        c = self._db_cursor
        try:
            test_case_row = next(c.execute("""SELECT * FROM cases WHERE number=?""", [index]))
        except StopIteration:
            raise exception.BoofuzzNoSuchTestCase()

        rows = c.execute("""SELECT * FROM steps WHERE test_case_index=?""", [index])
        steps = []
        for row in rows:
            data = row[3]
            # Little hack since BLOB becomes type buffer in py2 and bytes in py3
            # At the end, data will be equivalent types: bytes in py3 and str in py2
            try:
                if isinstance(data, buffer):
                    data = str(data)
            except NameError as e:
                if "buffer" in str(e):  # buffer type does not exist in py3
                    pass
                else:
                    raise
            steps.append(
                data_test_step.DataTestStep(
                    type=row[1], description=row[2], data=data, timestamp=row[4], truncated=row[5]
                )
            )
        return data_test_case.DataTestCase(
            name=test_case_row[0], index=test_case_row[1], timestamp=test_case_row[2], steps=steps
        )

    def query(self, query, params=None):
        if params is None:
            params = []
        c = self._db_cursor
        return c.execute(query, params)

    @property
    def failure_map(self):
        c = self._db_cursor
        failure_steps = c.execute('''SELECT * FROM steps WHERE type="fail"''')

        failure_map = collections.defaultdict(list)
        for step in failure_steps:
            failure_map[step[0]].append(step[2])
        return failure_map



import errno
import socket
import sys

from boofuzz import exception
from boofuzz.connections import base_socket_connection


class TCPSocketConnection(base_socket_connection.BaseSocketConnection):
    """BaseSocketConnection implementation for use with TCP Sockets.

    .. versionadded:: 0.2.0

    Args:
        host (str): Hostname or IP adress of target system.
        port (int): Port of target service.
        send_timeout (float): Seconds to wait for send before timing out. Default 5.0.
        recv_timeout (float): Seconds to wait for recv before timing out. Default 5.0.
        server (bool): Set to True to enable server side fuzzing.

    """

    def __init__(self, host, port, send_timeout=5.0, recv_timeout=5.0, server=False):
        super(TCPSocketConnection, self).__init__(send_timeout, recv_timeout)

        self.host = host
        self.port = port
        self.server = server
        self._serverSock = None

    def close(self):
        super(TCPSocketConnection, self).close()

        if self.server:
            self._serverSock.close()

    def open(self):
        self._open_socket()
        self._connect_socket()

    def _open_socket(self):
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        # call superclass to set timeout sockopt
        super(TCPSocketConnection, self).open()

    def _connect_socket(self):
        if self.server:
            self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            try:
                self._sock.bind((self.host, self.port))
            except socket.error as e:
                if e.errno == errno.EADDRINUSE:
                    raise exception.BoofuzzOutOfAvailableSockets()
                else:
                    raise

            self._serverSock = self._sock
            try:
                self._serverSock.listen(1)
                self._sock, addr = self._serverSock.accept()
            except socket.error as e:
                # When connection timeout expires, tear down the server socket so we can re-open it again after
                # restarting the target.
                self.close()
                if e.errno in [errno.EAGAIN]:
                    raise exception.BoofuzzTargetConnectionFailedError(str(e))
                else:
                    raise
        else:
            try:
                self._sock.connect((self.host, self.port))
            except socket.error as e:
                if e.errno == errno.EADDRINUSE:
                    raise exception.BoofuzzOutOfAvailableSockets()
                elif e.errno in [errno.ECONNREFUSED, errno.EINPROGRESS, errno.ETIMEDOUT]:
                    raise exception.BoofuzzTargetConnectionFailedError(str(e))
                else:
                    raise

    def recv(self, max_bytes):
        """
        Receive up to max_bytes data from the target.

        Args:
            max_bytes (int): Maximum number of bytes to receive.

        Returns:
            Received data.
        """
        data = b""

        try:
            data = self._sock.recv(max_bytes)
        except socket.timeout:
            data = b""
        except socket.error as e:
            if e.errno == errno.ECONNABORTED:
                raise exception.BoofuzzTargetConnectionAborted(
                    socket_errno=e.errno, socket_errmsg=e.strerror
                ).with_traceback(sys.exc_info()[2])
            elif (e.errno == errno.ECONNRESET) or (e.errno == errno.ENETRESET) or (e.errno == errno.ETIMEDOUT):
                raise exception.BoofuzzTargetConnectionReset().with_traceback(sys.exc_info()[2])
            elif e.errno == errno.EWOULDBLOCK:  # timeout condition if using SO_RCVTIMEO or SO_SNDTIMEO
                data = b""
            else:
                raise

        return data

    def send(self, data):
        """
        Send data to the target. Only valid after calling open!

        Args:
            data: Data to send.

        Returns:
            int: Number of bytes actually sent.
        """
        num_sent = 0

        try:
            num_sent = self._sock.send(data)
        except socket.error as e:
            if e.errno == errno.ECONNABORTED:
                raise exception.BoofuzzTargetConnectionAborted(
                    socket_errno=e.errno, socket_errmsg=e.strerror
                ).with_traceback(sys.exc_info()[2])
            elif e.errno in [errno.ECONNRESET, errno.ENETRESET, errno.ETIMEDOUT, errno.EPIPE]:
                raise exception.BoofuzzTargetConnectionReset().with_traceback(sys.exc_info()[2])
            else:
                raise

        return num_sent

    @property
    def info(self):
        return "{0}:{1}".format(self.host, self.port)



from .. import helpers
from ..fuzzable import Fuzzable


class Static(Fuzzable):
    """Static primitives are fixed and not mutated while fuzzing.

    :type name: str, optional
    :param name: Name, for referencing later. Names should always be provided, but if not, a default name will be given,
        defaults to None
    :type default_value: Raw, optional
    :param default_value: Raw static data
    """

    def __init__(self, name=None, default_value=None, *args, **kwargs):
        super(Static, self).__init__(name=name, default_value=default_value, fuzzable=False, *args, **kwargs)

    def encode(self, value, mutation_context):
        if value is None:
            value = b""
        return helpers.str_to_bytes(value)



#!/usr/bin/env python3
# Designed for use with boofuzz v0.3.0
#
# Autoprogramming an Internet Of Things (IOT) device by sending it a formatted xml data file with internal
# checksum over selected portions of the data.
# A typical CRC 16 algorithm is used to manage data integrity.
#
# how to run: python autoprog.py -a <target ip address> -p <port number>
#


import getopt  # command line arguments
import platform  # For getting the operating system name
import subprocess  # For executing a shell command
import sys
import time
from boofuzz import *
from crc16ccitt import getCrc16Ccitt

mylogger = FuzzLoggerText()
g_target_ip_addr = None
g_autoprogram_port = 65534


# Verify if the IOT target is still alive by expecting a response to a ping. This means that the test must have
# network access to the same subnet as the IOT device target. Verify that a ping reply is successful independent of
# running this fuzz test.
#
# noinspection PyMethodOverriding
# noinspection PyMethodParameters
class IOT_TargetMonitor(NetworkMonitor):
    def alive():
        global g_target_ip_addr
        param = "-n" if platform.system().lower() == "windows" else "-c"
        # One ping only
        command = ["ping", param, "1", g_target_ip_addr]
        # noinspection PyTypeChecker
        message = "alive() sending a ping command to " + g_target_ip_addr
        mylogger.log_info(message)
        try:
            subprocess.run(command, timeout=3)
        except subprocess.TimeoutExpired:
            return False
        else:
            mylogger.log_info("PING success")
            return True

    def pre_send(target=None, fuzz_data_logger=None, session=None, **kwargs):
        return

    def post_send(target=None, fuzz_data_logger=None, session=None, **kwargs):
        return True

    def retrieve_data():
        return

    def start_target():
        return True

    def set_options(*args, **kwargs):
        return

    def get_crash_synopsis():
        return "get_crash_synopsis detected a crash of the target."

    # The use of a 12 second sleep is based on experimentation for a specific IOT device. Change the number of seconds
    # as needed for your environment.
    def restart_target(target=None, **kwargs):
        mylogger.log_info("restart_target sleep for 12")
        time.sleep(12)
        if IOT_TargetMonitor.alive() is True:
            mylogger.log_info("restart_target ok")
            return True
        else:
            mylogger.log_info("restart_target failed")
            return False

    def post_start_target(target=None, fuzz_data_logger=None, session=None, **kwargs):
        return


def main(argv):
    # parse command line options.
    target_ip_addr = None
    autoprogram_port = 65534
    start_index = 1
    end_index = 65534

    global g_target_ip_addr
    global g_autoprogram_port

    try:
        opts, args = getopt.getopt(argv, "ha:p:s:e:", ["address=", "port=", "start_index=", "end_index="])
    except getopt.GetoptError:
        print(
            "autoprog.py --address|-a <target ip address> --port|-p <auto programming port> --start_index|\
            -s <start of fuzzing index> --end_index|-e <end of fuzzing index>"
        )
        sys.exit(2)
    for opt, arg in opts:
        if opt == "-h":
            print(
                "autoprog.py --address|-a <target ip address> --port|-p <auto programming port> --start_index|\
                -s <start of fuzzing index> --end_index|-e <end of fuzzing index>"
            )
            sys.exit()
        elif opt in ("-a", "--address"):
            target_ip_addr = arg
        elif opt in ("-p", "--port"):
            autoprogram_port = arg
        elif opt in ("-s", "--startindex"):
            start_index = arg
        elif opt in ("-e", "--endindex"):
            end_index = arg

    g_target_ip_addr = target_ip_addr
    g_autoprogram_port = int(autoprogram_port)
    target_message = "Target device ip address and port " + str(g_target_ip_addr) + " " + str(g_autoprogram_port)
    mylogger.log_info(target_message)

    IOT_TargetMonitor(host=g_target_ip_addr, port=g_autoprogram_port)

    mylogger.log_info("Initializing session to target ")
    session = Session(
        target=Target(
            connection=TCPSocketConnection(host=g_target_ip_addr, port=g_autoprogram_port),
            monitors=[IOT_TargetMonitor],
        ),
        # The use of a 12 second sleep is based on experimentation for a specific IOT device. Change the sleep count
        # as needed for your environment.
        sleep_time=12,
        crash_threshold_request=2,
        crash_threshold_element=2,
        index_start=int(start_index),
        index_end=int(end_index),
    )
    define_autoprog_static(session=session)
    mylogger.log_info("start fuzzing")

    session.fuzz()


#################################################################
# An Autoprogramming proposal protocol example
#################################################################
# Example protocol which will send the following data in sequence:
#    1. A URL inside a formatted line
#    2. The checksum of line 1.
#    3. A formatted line to indicate the completion of the checksum.
#    4. A formatted line indicating the start of the outer checksum.
#    5. An outer checksum managing data consisting of lines 1,2,3, and 4.
#    6. A formatted line to indicate the completion of the outer checksum.
#
# <ProposalList xmlns="http://www.example.com/HC/AutoProgramming">
# the_calculated_checksum_value_of_previous_line
# </Checksum_section_end>
# <ChecksumTotal_begins>
# the_calculated_outer_checksum_value
# </ChecksumTotal_ends>
#
# An example of an AutoProgram protocol output:
#    <ProposalList xmlns="http://www.example.com/HC/AutoProgramming">
#    1234
#    </Checksum_section_end>
#    <ChecksumTotal_begins>
#    5678
#    </ChecksumTotal_ends>
#
#################################################################


def define_autoprog_static(session):
    dl_line_1 = String(
        name="proposal_header", default_value='<ProposalList xmlns="http://www.mycompany.com/HC/AutoProgramming">'
    )
    # insert inner crc here
    dl_line_2 = String(name="crc_end", default_value="</Checksum_end>", fuzzable=False)
    dl_line_3 = String(name="outer_crc", default_value="<ChecksumTotal_begins>", fuzzable=False)
    # insert outer crc here
    dl_line_4 = String(name="outer_crcend", default_value="</ChecksumTotal_ends>", fuzzable=False)

    reqW = Request("autoprog")
    block = Block(name="autoprogB", request=reqW)
    reqW.push(block)
    crcValue_inner = Checksum(
        name="firstCRC16", block_name="autoprogB", request=reqW, algorithm=getCrc16Ccitt, length=2, fuzzable=False
    )
    crcValue_outer = Checksum(
        name="CRC16_outer", block_name="autoprogB", request=reqW, algorithm=getCrc16Ccitt, length=2, fuzzable=False
    )

    block.push(dl_line_1)
    reqW.push(crcValue_inner)
    block.push(dl_line_2)
    block.push(dl_line_3)
    reqW.push(crcValue_outer)
    block.push(dl_line_4)
    reqW.pop()

    session.connect(reqW)


if __name__ == "__main__":
    main(sys.argv[1:])



from .. import helpers
from ..fuzzable import Fuzzable
from ..protocol_session_reference import ProtocolSessionReference


class Repeat(Fuzzable):
    """Repeat the rendered contents of the specified block cycling from min_reps to max_reps counting by step.

    By default renders to nothing. This block modifier is useful for fuzzing overflows in table entries. This block
    modifier MUST come after the block it is being applied to.

    :param name: Name, for referencing later. Names should always be provided, but if not, a default name will be given,
        defaults to None
    :type name: str, optional
    :param block_name: Name of block to repeat
    :type block_name: str, optional
    :param request: Request this block belongs to, defaults to None
    :type request: boofuzz.Request, optional
    :param min_reps: Minimum number of block repetitions, defaults to 0
    :type min_reps: int, optional
    :param max_reps: Maximum number of block repetitions, defaults to None
    :type max_reps: int, optional
    :param step: Step count between min and max reps, defaults to 1
    :type step: int, optional
    :param variable: Repetitions will be derived from this variable, disables fuzzing, defaults to None
    :type variable: Boofuzz Integer Primitive, optional
    :param default_value: Value used when the element is not being fuzzed - should typically represent a valid value,
        defaults to None
    :type default_value: Raw
    :param fuzzable: Enable/disable fuzzing of this block, defaults to true
    :type fuzzable: bool, optional
    """

    def __init__(
        self,
        name=None,
        block_name=None,
        request=None,
        min_reps=0,
        max_reps=25,
        step=1,
        variable=None,
        default_value=None,
        *args,
        **kwargs
    ):
        if default_value is None:
            if variable is not None:
                default_value = ProtocolSessionReference(name=variable, default_value=0)
            else:
                default_value = 0

        super(Repeat, self).__init__(name=name, default_value=default_value, *args, **kwargs)

        self.block_name = block_name
        self.request = request
        self.min_reps = min_reps
        self.max_reps = max_reps
        self.step = step

        self._value = b""
        self._original_value = b""  # default to nothing!
        self._rendered = b""  # rendered value
        self._fuzz_complete = False  # flag if this primitive has been completely fuzzed
        self._fuzz_library = []  # library of static fuzz heuristics to cycle through.
        self._mutant_index = 0  # current mutation number
        self.current_reps = min_reps  # current number of repetitions

        if self.max_reps is not None and self.request is not None and self.block_name is not None:
            self._fuzz_library = list(range(self.min_reps, self.max_reps + 1, self.step))

    def mutations(self, default_value):
        for fuzzed_reps_number in self._fuzz_library:
            yield fuzzed_reps_number

    def num_mutations(self, default_value):
        """
        Determine the number of repetitions we will be making.

        @rtype:  int
        @return: Number of mutated forms this primitive can take.
        :param default_value:
        """
        return len(self._fuzz_library)

    def encode(self, value, mutation_context):
        return value * self._get_child_data(mutation_context=mutation_context)

    def _get_child_data(self, mutation_context):
        if self.request is not None and self.block_name is not None:
            _rendered = self.request.resolve_name(self.context_path, self.block_name).render(
                mutation_context=mutation_context
            )
        else:
            _rendered = ""
        return helpers.str_to_bytes(_rendered)

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self._name)

    def __len__(self):
        return (
            self.current_reps * len(self.request.names[self.block_name])
            if self.block_name is not None and self.request is not None
            else 0
        )



import atexit
import sys
import time
import warnings

from io import StringIO

try:
    import curses  # pytype: disable=import-error
except ImportError:
    # Allow fuzz_logger_curses to be imported on Windows -- will fail if you try to use it.
    warnings.warn("Importing curses failed. Optional console GUI features will not be available.", UserWarning)
    curses = None

import signal
import threading

from math import *
from shutil import get_terminal_size
from . import helpers
from . import ifuzz_logger_backend

COLOR_PAIR_WHITE = 1
COLOR_PAIR_CYAN = 2
COLOR_PAIR_RED = 3
COLOR_PAIR_YELLOW = 4
COLOR_PAIR_GREEN = 5
COLOR_PAIR_MAGENTA = 6
COLOR_PAIR_BLACK = 7

STATUS_RUNNING = 0
STATUS_PAUSED = 1
STATUS_DONE = 2


class FuzzLoggerCurses(ifuzz_logger_backend.IFuzzLoggerBackend):
    """
    This class formats FuzzLogger data for a console GUI using curses. This hasn't been tested on Windows.
    """

    INDENT_SIZE = 2

    def __init__(
        self,
        web_port=26000,
        web_address="localhost",
        window_height=40,
        window_width=130,
        auto_scroll=True,
        max_log_lines=500,
        wait_on_quit=True,
        min_refresh_rate=1000,
        bytes_to_str=helpers.hex_to_hexstr,
    ):
        """
        :type web_port: int
        :param web_port: Webinterface port. Default 26000

        :type window_height: int
        :param window_height: Default console height, set to on startup. Default 40

        :type window_width: int
        :param window_width: Default console width, set to on startup. Default 130

        :type auto_scroll: bool
        :param auto_scroll: Whether to auto-scroll the cases and crashed windows to always display the last line if
                            there are too many lines to display all of them. Default True

        :type max_log_lines: int
        :param max_log_lines: Maximum log lines to keep in the internal storage. Additional lines exceeding this limit
                              will not be displayed. Default 500

        :type wait_on_quit: bool
        :param wait_on_quit: Whether to keep the GUI open and wait for user-input when the main thread is about to exit.
                             Default True

        :type min_refresh_rate: int
        :param min_refresh_rate: The delay between two checks for a resize of the terminal in milliseconds.
                                 Increment 100 ms. Default 1000 ms

        :type bytes_to_str: function
        :param bytes_to_str: Function that converts sent/received bytes data to string for logging.
        """

        self._title = "boofuzz"
        self._web_port = web_port
        self._web_address = web_address
        self._max_log_lines = max_log_lines
        self._auto_scroll = auto_scroll
        self._current_data = None
        self._log_storage = []
        self._fail_storage = []
        self._wait_on_quit = wait_on_quit
        self._quit = False
        self._status = STATUS_RUNNING
        self._refresh_interval = min_refresh_rate
        self._event_resize = True
        self._event_log = False
        self._event_case_close = False
        self._event_crash = False

        self._total_index = 0
        self._total_num_mutations = 0
        self._current_name = ""
        self._current_index = 0
        self._current_num_mutations = 0

        self._format_raw_bytes = bytes_to_str
        self._version = helpers.get_boofuzz_version()

        # Resize console to minimum size
        self._width, self._height = get_terminal_size()
        if self._height < window_height or self._width < window_width:
            print("\x1b[8;{};{}t".format(window_height, window_width))
            self._height, self._width = window_height, window_width
        self._height_old = 0
        self._width_old = 0
        self._min_size_ok = True

        sys.stdout = sys.stderr = self._std_buffer = StringIO()
        atexit.register(self._cleanup)

        self._stdscr = curses.initscr()
        curses.start_color()
        curses.use_default_colors()
        curses.noecho()
        curses.curs_set(0)
        self._stdscr.nodelay(True)

        # Curses color pairs
        curses.init_pair(COLOR_PAIR_WHITE, curses.COLOR_WHITE, -1)
        curses.init_pair(COLOR_PAIR_CYAN, curses.COLOR_CYAN, -1)
        curses.init_pair(COLOR_PAIR_RED, curses.COLOR_RED, -1)
        curses.init_pair(COLOR_PAIR_YELLOW, curses.COLOR_YELLOW, -1)
        curses.init_pair(COLOR_PAIR_GREEN, curses.COLOR_GREEN, -1)
        curses.init_pair(COLOR_PAIR_MAGENTA, curses.COLOR_MAGENTA, -1)
        curses.init_pair(COLOR_PAIR_BLACK, curses.COLOR_BLACK, curses.COLOR_WHITE)

        # Start thread and restore the original SIGWINCH handler
        self._draw_thread = threading.Thread(name="curses_logger", target=self._draw_screen)
        self._draw_thread.setDaemon(True)
        current_signal_handler = signal.getsignal(signal.SIGWINCH)
        self._draw_thread.start()
        signal.signal(signal.SIGWINCH, current_signal_handler)

    def _cleanup(self):
        self._wait_on_quit = False
        self.close_test()
        sys.stderr = sys.__stderr__
        sys.stdout = sys.__stdout__
        print(self._std_buffer.getvalue())
        self._std_buffer.close()

    def open_test_case(self, test_case_id, name, index, *args, **kwargs):
        self._log_storage = []
        self._current_name = name
        self._total_index = index
        if "current_index" in kwargs:
            self._current_index = kwargs["current_index"]
        if "current_num_mutations" in kwargs:
            self._current_num_mutations = kwargs["current_num_mutations"]
        if "num_mutations" in kwargs:
            self._total_num_mutations = kwargs["num_mutations"]

        self._log_storage.append(
            helpers.format_log_msg(msg_type="test_case", description=test_case_id, format_type="curses")
        )
        self._event_log = True

    def open_test_step(self, description):
        self._log_storage.append(helpers.format_log_msg(msg_type="step", description=description, format_type="curses"))
        self._event_log = True

    def log_info(self, description):
        self._log_storage.append(helpers.format_log_msg(msg_type="info", description=description, format_type="curses"))
        self._event_log = True

    def log_check(self, description):
        self._log_storage.append(
            helpers.format_log_msg(msg_type="check", description=description, format_type="curses")
        )
        self._event_log = True

    def log_pass(self, description=""):
        self._log_storage.append(helpers.format_log_msg(msg_type="pass", description=description, format_type="curses"))
        self._event_log = True

    def log_fail(self, description="", indent_size=INDENT_SIZE):
        # TODO: Why do some fail messages have a trailing whitespace?
        fail_msg = (
            "#"
            + str(self._total_index)
            + (4 * indent_size + 1 - len(str(self._total_index))) * " "
            + description.strip()
        )
        self._fail_storage.append([fail_msg.replace("\n", " "), COLOR_PAIR_WHITE])
        self._log_storage.append(helpers.format_log_msg(msg_type="fail", description=description, format_type="curses"))
        self._event_crash = True
        self._event_log = True

    def log_error(self, description="", indent_size=INDENT_SIZE):
        fail_msg = (
            "#"
            + str(self._total_index)
            + (4 * indent_size + 1 - len(str(self._total_index))) * " "
            + description.strip()
        )
        self._fail_storage.append([fail_msg.replace("\n", " "), COLOR_PAIR_RED])
        self._log_storage.append(
            helpers.format_log_msg(msg_type="error", description=description, format_type="curses")
        )
        self._event_crash = True
        self._event_log = True

    def log_recv(self, data):
        self._log_storage.append(helpers.format_log_msg(msg_type="receive", data=data, format_type="curses"))
        self._event_log = True

    def log_send(self, data):
        self._log_storage.append(helpers.format_log_msg(msg_type="send", data=data, format_type="curses"))
        self._event_log = True

    def close_test_case(self):
        self._event_case_close = True

    def close_test(self):
        self._status = STATUS_DONE
        self._quit = True
        try:
            if self._wait_on_quit:
                self._draw_thread.join()
            else:
                self._draw_thread.join(max(0.2, self._refresh_interval * 3))
        except KeyboardInterrupt:
            pass
        finally:
            self._wait_on_quit = False
            curses.nocbreak()
            self._stdscr.keypad(False)
            curses.echo()
            curses.endwin()

    def _draw_main(self, force=False):
        self._width, self._height = get_terminal_size()
        if not force and self._width == self._width_old and self._height == self._height_old:
            return
        self._height_old = self._height
        self._width_old = self._width
        curses.resizeterm(self._height, self._width)
        self._stdscr.erase()
        if self._height < 40 or self._width < 130:
            self._min_size_ok = False
            return
        else:
            self._min_size_ok = True

        # Render title
        self._stdscr.addstr(0, 0, "=" * self._width)
        start_x_title = int((self._width // 2) - (len(self._title) // 2) - len(self._title) % 2)
        self._stdscr.addstr(0, start_x_title, self._title, curses.color_pair(COLOR_PAIR_CYAN) | curses.A_BOLD)

        # Render status bar
        self._stdscr.attron(curses.color_pair(COLOR_PAIR_BLACK))
        self._stdscr.addstr(self._height - 1, 0, " " * (self._width - 1))
        self._stdscr.insch(" ")  # Fill bottom right corner
        if self._quit:
            self._stdscr.addstr(
                self._height - 1, 1, "Press 'q' to quit", curses.color_pair(COLOR_PAIR_BLACK) | curses.A_BLINK
            )
        else:
            self._stdscr.addstr(self._height - 1, 1, "Press 'CTRL+C' to abort")
        self._stdscr.addstr(self._height - 1, self._width - len(self._version) - 1, self._version)
        self._stdscr.attroff(curses.color_pair(COLOR_PAIR_BLACK))
        self._stdscr.refresh()

        # Initialise test case window
        self._casescr_frame = curses.newpad(self._max_log_lines + 1, self._width)
        self._casescr_frame.nodelay(True)
        self._casescr_frame.border()
        self._casescr_frame.addstr(0, 1, "Test case log", curses.color_pair(COLOR_PAIR_YELLOW) | curses.A_BOLD)
        self._casescr_frame.refresh(0, 0, 1, 0, self._height - 18, self._width)
        self._casescr = self._casescr_frame.subpad(self._max_log_lines, self._width - 2, 1, 1)
        self._draw_case()

        # Initialise crash window
        self._crashescr_frame = curses.newpad(self._max_log_lines + 1, self._width)
        self._crashescr_frame.nodelay(True)
        self._crashescr_frame.border()
        self._crashescr_frame.addstr(0, 1, "Crashes", curses.color_pair(COLOR_PAIR_RED) | curses.A_BOLD)
        self._crashescr_frame.refresh(0, 0, self._height - 17, 0, self._height - 8, self._width)
        self._crashescr = self._crashescr_frame.subpad(self._max_log_lines, self._width - 2, 1, 1)
        self._draw_crash()

        # Initialise status window
        self._statscr = curses.newwin(6, self._width, self._height - 7, 0)
        self._statscr.nodelay(True)
        self._statscr.border()
        self._statscr.addstr(0, 1, "Status", curses.color_pair(COLOR_PAIR_CYAN) | curses.A_BOLD)
        self._draw_stat()

    def _draw_case(self, indent_size=INDENT_SIZE):
        # Test Case Screen
        # TODO: Handle longer indent for multi-line 'fail' messages
        self._casescr.erase()
        total_indent_size = indent_size * 2 + 1 + 25

        _render_pad(
            lines=self._log_storage[: self._max_log_lines],
            pad=self._casescr,
            y_min=2,
            x_min=1,
            y_max=self._height - 18,
            x_max=self._width - 1,
            max_lines=self._max_log_lines,
            total_indent_size=total_indent_size,
            auto_scroll=self._auto_scroll,
        )

    def _draw_crash(self, indent_size=INDENT_SIZE):
        # Crashes Screen
        total_indent_size = indent_size * 5

        _render_pad(
            lines=self._fail_storage[: self._max_log_lines],
            pad=self._crashescr,
            y_min=self._height - 16,
            x_min=1,
            y_max=self._height - 8,
            x_max=self._width - 1,
            max_lines=self._max_log_lines,
            total_indent_size=total_indent_size,
            auto_scroll=self._auto_scroll,
            truncate_long_lines=True,
        )

    def _draw_stat(self):
        # Status Screen
        self._indent_size = 16
        self._statscr.addstr(1, 1, "Webinterface:")
        self._statscr.addstr(1, self._indent_size, "{}:{}".format(self._web_address, self._web_port))
        self._statscr.addstr(2, 1, "Case:")
        # fmt: off
        self._statscr.addstr(2, self._indent_size, _progess_bar(self._current_index,
                                                                self._current_num_mutations,
                                                                self._width - self._indent_size))
        self._statscr.addstr(3, 1, "Total:")
        self._statscr.addstr(3, self._indent_size, _progess_bar(self._total_index,
                                                                self._total_num_mutations,
                                                                self._width - self._indent_size))
        # fmt: on
        # TODO: Get paused flag from sessions
        if self._status == STATUS_RUNNING:
            self._statscr.addstr(4, 1, "Status:")
            self._statscr.addstr(4, self._indent_size, "Running", curses.color_pair(COLOR_PAIR_YELLOW))
        elif self._status == STATUS_PAUSED:
            self._statscr.addstr(4, 1, "Status:")
            self._statscr.addstr(4, self._indent_size, "Paused ", curses.color_pair(COLOR_PAIR_RED) | curses.A_BLINK)
        elif self._status == STATUS_DONE:
            self._statscr.addstr(4, 1, "Status:")
            self._statscr.addstr(4, self._indent_size, "Done   ", curses.color_pair(COLOR_PAIR_GREEN))

        self._statscr.refresh()

    def _draw_screen(self):
        error_counter = 0
        ms_since_refresh = 0
        key = 0
        wait_for_key = False
        try:
            while not ((key == ord("q") or not self._wait_on_quit) and self._quit):
                try:
                    if self._event_resize or ms_since_refresh >= self._refresh_interval:
                        self._draw_main()
                        ms_since_refresh = 0
                        self._event_resize = False

                    if self._quit and not wait_for_key:
                        self._draw_main(force=True)
                        wait_for_key = True

                    if self._min_size_ok:
                        if self._event_log:
                            self._draw_case()
                            self._event_log = False

                        if self._event_crash:
                            self._draw_crash()
                            self._event_crash = False

                        if self._event_case_close:
                            self._draw_stat()
                            self._event_case_close = False

                    key = self._stdscr.getch()
                    curses.flushinp()

                    time.sleep(0.1)
                    ms_since_refresh += 100
                    error_counter = 0
                except curses.error:
                    error_counter += 1
                    if error_counter > 2:
                        raise
        finally:
            curses.nocbreak()
            self._stdscr.keypad(False)
            curses.echo()
            curses.endwin()


def _progess_bar(current, total, width):
    try:
        percent = current / total
    except (ZeroDivisionError, TypeError):
        percent = 0
        total = 0
    title_str = "{:7d} of {:7d} ".format(current, total)
    percent_str = " {:7.3f}%".format(percent * 100)
    bar_len = width - 4 - len(title_str) - len(percent_str)
    num_bars = int(round(percent * bar_len))
    bar_str = "[" + "=" * num_bars + " " * (bar_len - num_bars) + "]"
    return title_str + bar_str + percent_str


def _render_pad(
    lines, pad, y_min, x_min, y_max, x_max, max_lines, total_indent_size, auto_scroll, truncate_long_lines=False
):
    total_rows = 0
    height = y_max - y_min + 1
    width = x_max - x_min

    for i in range(len(lines)):
        if total_rows < max_lines - 1:
            pad.addnstr(total_rows, 0, lines[i][0], width, curses.color_pair(lines[i][1]))
            total_rows += 1
        else:
            pad.addstr(
                total_rows,
                0,
                "Maximum number of lines reached for this window! Increase 'max_log_lines'",
                curses.color_pair(COLOR_PAIR_RED),
            )
            total_rows += 1
            break

        if not truncate_long_lines:
            columns = width - total_indent_size
            rows = int(ceil(len(lines[i][0][width:]) / columns))
            if rows >= 1:
                for row in range(1, rows + 1):
                    if total_rows < max_lines - 1:
                        pad.addstr(
                            total_rows,
                            total_indent_size,
                            lines[i][0][width:][(row * columns) - columns : row * columns],
                            curses.color_pair(lines[i][1]),
                        )
                        total_rows += 1
                    else:
                        break

    if total_rows > height and auto_scroll:
        offset = total_rows - height
        pad.refresh(offset, 0, y_min, x_min, y_max, x_max)
    else:
        pad.refresh(0, 0, y_min, x_min, y_max, x_max)



# MSRPC NDR TYPES
import struct

from .. import blocks, exception, helpers, primitives
from ..helpers import calculate_four_byte_padding


class NdrConformantArray(blocks.Block):
    """
    Note: this is not for fuzzing the RPC protocol but rather just representing an NDR string for fuzzing the actual
    client.
    """

    def __init__(self, name, request, value, options=None):
        if not options:
            options = {}

        super(NdrConformantArray).__init__(name, request)

        self.value = value
        self.options = options

        if not self.value:
            raise exception.SullyRuntimeError("MISSING LEGO.ndr_conformant_array DEFAULT VALUE")

        self.push(primitives.String())

    def render(self, mutation_context=None):
        """
        We overload and extend the render routine in order to properly pad and prefix the string.

        [dword length][array][pad]
        """

        # let the parent do the initial render.
        blocks.Block.render(self)

        # encode the empty string correctly:
        if self._rendered == b"":
            self._rendered = b"\x00\x00\x00\x00"
        else:
            string_with_padding = self._rendered + calculate_four_byte_padding(self._rendered)
            self._rendered = struct.pack("<L", len(self._rendered)) + string_with_padding

        return helpers.str_to_bytes(self._rendered)


class NdrString(blocks.Block):
    """
    Note: this is not for fuzzing the RPC protocol but rather just representing an NDR string for fuzzing the actual
    client.
    """

    def __init__(self, name, request, value, options=None):
        if not options:
            options = {}

        super(NdrString, self).__init__(name, request)

        self.value = value
        self.options = options

        if not self.value:
            raise exception.SullyRuntimeError("MISSING LEGO.tag DEFAULT VALUE")

        self.push(primitives.String(name=name + "_STR", default_value=""))

    def render(self):
        """
        We overload and extend the render routine in order to properly pad and prefix the string.

        [dword length][dword offset][dword passed size][string][pad]
        """

        # let the parent do the initial render.
        blocks.Block.render(self)

        # encode the empty string correctly:
        if self._rendered == b"":
            self._rendered = b"\x00\x00\x00\x00"
        else:
            # ensure null termination.
            self._rendered += b"\x00"

            # format accordingly.
            length = len(self._rendered)
            self._rendered = (
                b""
                + struct.pack("<L", length)
                + struct.pack("<L", 0)
                + struct.pack("<L", length)
                + self._rendered
                + calculate_four_byte_padding(self._rendered)
            )

        return helpers.str_to_bytes(self._rendered)


class NdrWString(blocks.Block):
    """
    Note: this is not for fuzzing the RPC protocol but rather just representing an NDR string for fuzzing the actual
    client.
    """

    def __init__(self, name, request, value, options=None):
        if not options:
            options = {}

        super(NdrWString).__init__(name, request)

        self.value = value
        self.options = options

        if not self.value:
            raise exception.SullyRuntimeError("MISSING LEGO.tag DEFAULT VALUE")

        self.push(primitives.String())

    def render(self):
        """
        We overload and extend the render routine in order to properly pad and prefix the string.

        [dword length][dword offset][dword passed size][string][pad]
        """

        # let the parent do the initial render.
        blocks.Block.render(self)

        # encode the empty string correctly:
        if self._rendered == b"":
            self._rendered = b"\x00\x00\x00\x00"
        else:
            # unicode encode and null terminate.
            self._rendered = self._rendered.encode("utf-16le") + b"\x00"

            # format accordingly.
            length = len(self._rendered)
            self._rendered = (
                b""
                + struct.pack("<L", length)
                + struct.pack("<L", 0)
                + struct.pack("<L", length)
                + self._rendered
                + calculate_four_byte_padding(self._rendered)
            )

        return helpers.str_to_bytes(self._rendered)



import math
import struct

from ..helpers import crc16


def dnp3(data, control_code=b"\x44", src=b"\x00\x00", dst=b"\x00\x00"):
    num_packets = int(math.ceil(float(len(data)) / 250.0))
    packets = []

    for i in range(num_packets):
        packet_slice = data[i * 250 : (i + 1) * 250]

        p = b"\x05\x64"
        p += len(packet_slice).to_bytes(1, "little")
        p += control_code
        p += dst
        p += src

        chksum = struct.pack("<H", crc16(p))

        p += chksum

        num_chunks = int(math.ceil(float(len(packet_slice) / 16.0)))

        # insert the fragmentation flags / sequence number.
        # first frag: 0x40, last frag: 0x80

        frag_number = i

        if i == 0:
            frag_number |= 0x40

        if i == num_packets - 1:
            frag_number |= 0x80

        p += frag_number.to_bytes(1, "little")

        for x in range(num_chunks):
            chunk = packet_slice[i * 16 : (i + 1) * 16]
            chksum = struct.pack("<H", crc16(chunk))
            p += chksum + chunk

        packets.append(p)

    return packets



import itertools

from boofuzz.mutation import Mutation
from .mutation_context import MutationContext
from .protocol_session_reference import ProtocolSessionReference


class Fuzzable:
    """Parent class for all primitives and blocks.

    When making new fuzzable types, one will typically override :meth:`mutations` and/or :meth:`encode`.

    :meth:`mutations` is a generator function yielding mutations, typically of type bytes.

    :meth:`encode` is a function that takes a value and encodes it. The value comes from :meth:`mutations` or
    ``default_value``. :class:`FuzzableBlock <boofuzz.FuzzableBlock>` types can also encode the data generated by
    child nodes.

    Implementors may also want to override :meth:`num_mutations` -- the default implementation manually exhausts
    :meth:`mutations` to get a number.

    The rest of the methods are used by boofuzz to handle fuzzing and are typically not overridden.

    :type name: str, optional
    :param name: Name, for referencing later. Names should always be provided, but if not, a default name will be given,
        defaults to None
    :type default_value: Any, optional
    :param default_value: Value used when the element is not being fuzzed - should typically represent a valid value.
        Can be a static value, or a ReferenceValueTestCaseSession, defaults to None
    :type fuzzable: bool, optional
    :param fuzzable: Enable fuzzing of this primitive, defaults to True
    :type fuzz_values: list, optional
    :param fuzz_values: List of custom fuzz values to add to the normal mutations, defaults to None
    """

    name_counter = 0

    def __init__(self, name=None, default_value=None, fuzzable=True, fuzz_values=None):
        self._fuzzable = fuzzable
        self._name = name
        self._default_value = default_value
        self._context_path = ""
        self._request = None
        self._halt_mutations = False
        if fuzz_values is None:
            fuzz_values = list()
        self._fuzz_values = fuzz_values

        if self._name is None:
            Fuzzable.name_counter += 1
            self._name = "{0}{1}".format(type(self).__name__, Fuzzable.name_counter)

    @property
    def fuzzable(self):
        """If False, this element should not be mutated in normal fuzzing."""
        return self._fuzzable

    @property
    def name(self):
        """Element name, should be unique for each instance.

        :rtype: str
        """
        return self._name

    @property
    def qualified_name(self):
        """Dot-delimited name that describes the request name and the path to the element within the request.

        Example: "request1.block1.block2.node1"

        """
        return ".".join(s for s in (self._context_path, self.name) if s != "")

    @property
    def context_path(self):
        """Dot-delimited string that describes the path up to this element. Configured after the object is attached
        to a Request."""
        if not hasattr(self, "_context_path"):
            self._context_path = None
        return self._context_path

    @context_path.setter
    def context_path(self, x):
        self._context_path = x

    @property
    def request(self):
        """Reference to the Request to which this object is attached."""
        if not hasattr(self, "_request"):
            self._request = None
        return self._request

    @request.setter
    def request(self, x):
        self._request = x

    def stop_mutations(self):
        """Stop yielding mutations on the currently running :py:meth:`mutations` call.

        Used by boofuzz to stop fuzzing an element when it's already caused several failures.

        Returns:
            NoneType: None
        """
        self._halt_mutations = True

    def original_value(self, test_case_context=None):
        """Original, non-mutated value of element.

        Args:
            test_case_context (ProtocolSession): Used to resolve ReferenceValueTestCaseSession type default values.

        Returns:
        """
        if isinstance(self._default_value, ProtocolSessionReference):
            if test_case_context is None:
                return self._default_value.default_value
            else:
                return test_case_context.session_variables[self._default_value.name]
        else:
            return self._default_value

    def get_mutations(self):
        """Iterate mutations. Used by boofuzz framework.

        Yields:
            list of Mutation: Mutations

        """
        try:
            if not self.fuzzable:
                return
            index = 0
            for value in itertools.chain(self.mutations(self.original_value()), self._fuzz_values):
                if self._halt_mutations:
                    self._halt_mutations = False
                    return
                if isinstance(value, list):
                    yield value
                elif isinstance(value, Mutation):
                    yield [value]
                else:
                    yield [Mutation(value=value, qualified_name=self.qualified_name, index=index)]
                    index += 1
        finally:
            self._halt_mutations = False  # in case stop_mutations is called when mutations were exhausted anyway

    def render(self, mutation_context=None):
        """Render after applying mutation, if applicable.
        :type mutation_context: MutationContext
        """
        return self.encode(value=self.get_value(mutation_context=mutation_context), mutation_context=mutation_context)

    def get_num_mutations(self):
        return self.num_mutations(default_value=self.original_value(test_case_context=None)) + len(self._fuzz_values)

    def get_value(self, mutation_context=None):
        """Helper method to get the currently applicable value.

        This is either the default value, or the active mutation value as dictated by mutation_context.

        Args:
            mutation_context (MutationContext):

        Returns:

        """
        if mutation_context is None:
            mutation_context = MutationContext()
        if self.qualified_name in mutation_context.mutations:
            mutation = mutation_context.mutations[self.qualified_name]
            if callable(mutation.value):
                value = mutation.value(self.original_value(test_case_context=mutation_context.protocol_session))
            else:
                value = mutation.value
        else:
            value = self.original_value(test_case_context=mutation_context.protocol_session)

        return value

    def mutations(self, default_value):
        """Generator to yield mutation values for this element.

        Values are either plain values or callable functions that take a "default value" and mutate it. Functions are
        used when the default or "normal" value influences the fuzzed value. Functions are used because the "normal"
        value is sometimes dynamic and not known at the time of generation.

        Each mutation should be a pre-rendered value. That is, it must be suitable to pass to encode().

        Default: Empty iterator.

        Args:
            default_value:
        """
        return
        yield

    def encode(self, value, mutation_context):
        """Takes a value and encodes/renders/serializes it to a bytes (byte string).

        Optional if mutations() yields bytes.

        Example: Yield strings with mutations() and encode them to UTF-8 using encode().

        Default behavior: Return value.

        Args:
            value: Value to encode. Type should match the type yielded by mutations()
            mutation_context (MutationContext): Context for current mutation, if any.


        Returns:
            bytes: Encoded/serialized value.
        """
        return value

    def num_mutations(self, default_value):
        """Return the total number of mutations for this element (not counting "fuzz_values").

        Default implementation exhausts the mutations() generator, which is inefficient. Override if you can provide a
        value more efficiently, or if exhausting the mutations() generator has side effects.

        Args:
            default_value: Use if number of mutations depends on the default value. Provided by FuzzableWrapper.
                Note: It is generally good behavior to have a consistent number of mutations for a given default value
                length.

        Returns:
            int: Number of mutated forms this primitive can take
        """
        return sum(1 for _ in self.mutations(default_value=default_value))

    def __repr__(self):
        return "<%s %s %s>" % (
            self.__class__.__name__,
            self.name,
            repr(self.original_value(test_case_context=None)),
        )

    def __len__(self):
        """Length of field. May vary if mutate() changes the length.

        Returns:
            int: Length of element (length of mutated element if mutated).
        """
        return len(self.render(MutationContext()))

    def __bool__(self):
        """Make sure instances evaluate to True even if __len__ is zero.

        Design Note: Exists in case some wise guy uses `if my_element:` to
        check for null value.

        Returns:
            bool: True
        """
        return True



#
# pGRAPH
# Copyright (C) 2006 Pedram Amini <pedram.amini@gmail.com>
#
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#


class Cluster:
    id = None
    nodes = []

    def __init__(self, cluster_id=None):
        """
        Class constructor.
        """

        self.id = cluster_id
        self.nodes = []

    def add_node(self, node):
        """
        Add a node to the cluster.

        @type  node: pGRAPH Node
        @param node: Node to add to cluster
        """

        self.nodes.append(node)

        return self

    def del_node(self, node_id):
        """
        Remove a node from the cluster.

        @type  node_id: pGRAPH Node
        @param node_id: Node to remove from cluster
        """

        for node in self.nodes:
            if node.id == node_id:
                self.nodes.remove(node)
                break

        return self

    def find_node(self, attribute, value):
        """
        Find and return the node with the specified attribute / value pair.

        @type  attribute: str
        @param attribute: Attribute name we are looking for
        @type  value:     Mixed
        @param value:     Value of attribute we are looking for

        @rtype:  Mixed
        @return: Node, if attribute / value pair is matched. None otherwise.
        """

        for node in self.nodes:
            if hasattr(node, attribute):
                if getattr(node, attribute) == value:
                    return node

        return None

    def render(self):
        pass



#!/usr/bin/env python3
# Designed for use with boofuzz v0.2.0
#
# Minimal example which fuzzes a remote HTTPS server
# Use this as a starting point for SSL/TLS server fuzzing

from boofuzz import *

# If you don't want to verify remote certificate, create a SSLContext.
# WARNING: You will be vulnerable to a man-in-the-middle attack!
#   import ssl
#   ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
#   ctx.check_hostname = False
#   ctx.verify_mode = ssl.CERT_NONE

session = Session(
    target=Target(
        connection=SSLSocketConnection(
            host="127.0.0.1",
            port=443,
            server_hostname="example.com",  # Hostname must match the remote certificate
            # sslcontext=ctx,
        )
    )
)
s_initialize("GET Request")
s_string("GET", name="request method")
s_delim(" ")
s_string("/", name="resource")
s_static("\r\n\r\n")
session.connect(session.root, s_get("GET Request"))
session.fuzz()



from .fuzzable import Fuzzable


class FuzzableBlock(Fuzzable):
    """Fuzzable type designed to have children elements.

    FuzzableBlock overrides the following methods, changing the default behavior for any type based on FuzzableBlock:

    1. :meth:`mutations` Iterate through the mutations yielded by all child nodes.
    2. :meth:`num_mutations` Sum the mutations represented by each child node.
    3. :meth:`encode` Call :meth:`get_child_data`.

    FuzzableBlock adds the following methods:

    1. :meth:`get_child_data` Render and concatenate all child nodes.
    2. :meth:`push` Add an additional child node; generally used only internally.

    :param name: Name, for referencing later. Names should always be provided, but if not, a default name will be given,
        defaults to None
    :type name: str, optional
    :param request: Request this block belongs to, defaults to None
    :type request: boofuzz.Request, optional
    :param children: List of child nodes (typically given to FuzzableBlock types)m defaults to None
    :type children: boofuzz.Fuzzable, optional
    """

    def __init__(self, name=None, request=None, children=None, *args, **kwargs):
        super(FuzzableBlock, self).__init__(name=name, *args, **kwargs)
        self.request = request

        if children is None:
            self.stack = []
        elif isinstance(children, Fuzzable):
            self.stack = [children]
        else:
            self.stack = list(children)

    def mutations(self, default_value, skip_elements=None):
        if skip_elements is None:
            skip_elements = []
        for item in self.stack:
            if item.qualified_name in skip_elements:
                continue
            self.request.mutant = item
            for mutation in item.get_mutations():
                yield mutation

    def num_mutations(self, default_value=None):
        num_mutations = 0

        for item in self.stack:
            if item.fuzzable:
                num_mutations += item.get_num_mutations()

        return num_mutations

    def get_child_data(self, mutation_context):
        """Get child or referenced data for this node.

        For blocks that reference other data from the message structure (e.g. size, checksum, blocks). See
        FuzzableBlock for an example.

        Args:
            mutation_context (MutationContext): Mutation context.

        Returns:
            bytes: Child data.
        """
        rendered = b""
        for item in self.stack:
            rendered += item.render(mutation_context=mutation_context)
        return rendered

    def encode(self, value, mutation_context):
        return self.get_child_data(mutation_context=mutation_context)

    def push(self, item):
        """Push a child element onto this block's stack.

        Args:
            item (Fuzzable): Some wrapped Fuzzable element

        Returns: None
        """
        self.stack.append(item)



#
# pGRAPH
# Copyright (C) 2006 Pedram Amini <pedram.amini@gmail.com>
#
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

import copy

import pydot


class Graph:
    """
    @todo: Add support for clusters
    @todo: Potentially swap node list with a node dictionary for increased performance
    """

    id = None
    clusters = []
    edges = {}
    nodes = {}

    def __init__(self, graph_id=None):
        self.id = graph_id
        self.clusters = []
        self.edges = {}
        self.nodes = {}

    def add_cluster(self, cluster):
        """
        Add a pgraph cluster to the graph.

        @type  cluster: pGRAPH Cluster
        @param cluster: Cluster to add to graph
        """

        self.clusters.append(cluster)

        return self

    def add_edge(self, graph_edge, prevent_dups=True):
        """
        Add a pgraph edge to the graph. Ensures a node exists for both the source and destination of the edge.

        @type  graph_edge:         pGRAPH Edge
        @param graph_edge:         Edge to add to graph
        @type  prevent_dups: Boolean
        @param prevent_dups: (Optional, Def=True) Flag controlling whether or not the addition of duplicate edges is ok
        """

        if prevent_dups:
            if graph_edge.id in self.edges:
                return self

        # ensure the source and destination nodes exist.
        if self.find_node("id", graph_edge.src) is not None and self.find_node("id", graph_edge.dst) is not None:
            self.edges[graph_edge.id] = graph_edge

        return self

    def add_graph(self, other_graph):
        """
        Alias of graph_cat(). Concatenate the other graph into the current one.

        @todo: Add support for clusters
        @see:  graph_cat()

        @type  other_graph: pgraph.Graph
        @param other_graph: Graph to concatenate into this one.
        """

        return self.graph_cat(other_graph)

    def add_node(self, node):
        """
        Add a pgraph node to the graph. Ensures a node with the same id does not already exist in the graph.

        @type  node: pGRAPH Node
        @param node: Node to add to graph
        """

        node.number = len(self.nodes)

        if node.id not in self.nodes:
            self.nodes[node.id] = node

        return self

    def del_cluster(self, cluster_id):
        """
        Remove a cluster from the graph.

        @type  cluster_id: Mixed
        @param cluster_id: Identifier of cluster to remove from graph
        """

        for cluster in self.clusters:
            if cluster.id == cluster_id:
                self.clusters.remove(cluster)
                break

        return self

    def del_edge(self, graph_id=None, src=None, dst=None):
        """
        Remove an edge from the graph. There are two ways to call this routine, with an edge id::

            graph.del_edge(id)

        or by specifying the edge source and destination::

            graph.del_edge(src=source, dst=destination)

        @type  graph_id:  Mixed
        @param graph_id:  (Optional) Identifier of edge to remove from graph
        @type  src:       Mixed
        @param src:       (Optional) Source of edge to remove from graph
        @type  dst:       Mixed
        @param dst:       (Optional) Destination of edge to remove from graph
        """

        if not graph_id:
            graph_id = (src << 32) + dst  # pytype: disable=unsupported-operands

        if graph_id in self.edges:
            del self.edges[graph_id]

        return self

    def del_graph(self, other_graph):
        """
        Alias of graph_sub(). Remove the elements shared between the current graph and other graph from the current
        graph.

        @todo: Add support for clusters
        @see:  graph_sub()

        @type  other_graph: pgraph.Graph
        @param other_graph: Graph to diff/remove against
        """

        return self.graph_sub(other_graph)

    def del_node(self, node_id):
        """
        Remove a node from the graph.

        @type  node_id: Mixed
        @param node_id: Identifier of node to remove from graph
        """

        if node_id in self.nodes:
            del self.nodes[node_id]

        return self

    def edges_from(self, edge_id):
        """
        Enumerate the edges from the specified node.

        @type  edge_id: Mixed
        @param edge_id: Identifier of node to enumerate edges from

        @rtype:  list
        @return: List of edges from the specified node
        """

        return [edge_value for edge_value in list(self.edges.values()) if edge_value.src == edge_id]

    def edges_to(self, edge_id):
        """
        Enumerate the edges to the specified node.

        @type  edge_id: Mixed
        @param edge_id: Identifier of node to enumerate edges to

        @rtype:  list
        @return: List of edges to the specified node
        """

        return [edge_value for edge_value in list(self.edges.values()) if edge_value.dst == edge_id]

    def find_cluster(self, attribute, value):
        """
        Find and return the cluster with the specified attribute / value pair.

        @type  attribute: str
        @param attribute: Attribute name we are looking for
        @type  value:     Mixed
        @param value:     Value of attribute we are looking for

        @rtype:  Mixed
        @return: Cluster, if attribute / value pair is matched. None otherwise.
        """

        for cluster in self.clusters:
            if hasattr(cluster, attribute):
                if getattr(cluster, attribute) == value:
                    return cluster

        return None

    def find_cluster_by_node(self, attribute, value):
        """
        Find and return the cluster that contains the node with the specified attribute / value pair.

        @type  attribute: str
        @param attribute: Attribute name we are looking for
        @type  value:     Mixed
        @param value:     Value of attribute we are looking for

        @rtype:  Mixed
        @return: Cluster, if node with attribute / value pair is matched. None otherwise.
        """

        for cluster in self.clusters:
            for node in cluster:
                if hasattr(node, attribute):
                    if getattr(node, attribute) == value:
                        return cluster

        return None

    def find_edge(self, attribute, value):
        """
        Find and return the edge with the specified attribute / value pair.

        @type  attribute: str
        @param attribute: Attribute name we are looking for
        @type  value:     Mixed
        @param value:     Value of attribute we are looking for

        @rtype:  Mixed
        @return: Edge, if attribute / value pair is matched. None otherwise.
        """

        # if the attribute to search for is the id, simply return the edge from the internal hash.
        if attribute == "id" and value in self.edges:
            return self.edges[value]

        # step through all the edges looking for the given attribute/value pair.
        else:
            # TODO: Verify that this actually works? Was broken when I got here ;-P
            for node_edge in list(self.edges.values()):
                if hasattr(node_edge, attribute):
                    if getattr(node_edge, attribute) == value:
                        return node_edge

        return None

    def find_node(self, attribute, value):
        """
        Find and return the node with the specified attribute / value pair.

        @type  attribute: str
        @param attribute: Attribute name we are looking for
        @type  value:     mixed
        @param value:     Value of attribute we are looking for

        @rtype:  Mixed
        @return: Node, if attribute / value pair is matched. None otherwise.
        """

        # if the attribute to search for is the id, simply return the node from the internal hash.
        if attribute == "id" and value in self.nodes:
            return self.nodes[value]

        # step through all the nodes looking for the given attribute/value pair.
        else:
            for node in list(self.nodes.values()):
                if hasattr(node, attribute):
                    if getattr(node, attribute) == value:
                        return node

        return None

    def graph_cat(self, other_graph):
        """
        Concatenate the other graph into the current one.

        @todo: Add support for clusters

        @type  other_graph: pgraph.Graph
        @param other_graph: Graph to concatenate into this one.
        """

        for other_node in list(other_graph.nodes.values()):
            self.add_node(other_node)

        for other_edge in list(other_graph.edges.values()):
            self.add_edge(other_edge)

        return self

    def graph_down(self, from_node_id, max_depth=-1):
        """
        Create a new graph, looking down, from the specified node id to the specified depth.

        @type  from_node_id: pgraph.node
        @param from_node_id: Node to use as start of down graph
        @type  max_depth:    Integer
        @param max_depth:    (Optional, Def=-1) Number of levels to include in down graph (-1 for infinite)

        @rtype:  pgraph.Graph
        @return: Down graph around specified node.
        """

        down_graph = Graph()
        from_node = self.find_node("id", from_node_id)

        if not from_node:
            print("unable to resolve node {:08x}".format(from_node_id))
            raise Exception

        levels_to_process = []
        current_depth = 1

        levels_to_process.append([from_node])

        for level in levels_to_process:
            next_level = []

            # noinspection PyChainedComparisons
            if current_depth > max_depth and max_depth != -1:
                break

            for node in level:
                down_graph.add_node(copy.copy(node))

                for edge in self.edges_from(node.id):
                    to_add = self.find_node("id", edge.dst)

                    if not down_graph.find_node("id", edge.dst):
                        next_level.append(to_add)

                    down_graph.add_node(copy.copy(to_add))
                    down_graph.add_edge(copy.copy(edge))

            if next_level:
                levels_to_process.append(next_level)

            current_depth += 1

        return down_graph

    def graph_intersect(self, other_graph):
        """
        Remove all elements from the current graph that do not exist in the other graph.

        @todo: Add support for clusters

        @type  other_graph: pgraph.Graph
        @param other_graph: Graph to intersect with
        """

        for node in list(self.nodes.values()):
            if not other_graph.find_node("id", node.id):
                self.del_node(node.id)

        for edge in list(self.edges.values()):
            if not other_graph.find_edge("id", edge.id):
                self.del_edge(edge.id)

        return self

    def graph_proximity(self, center_node_id, max_depth_up=2, max_depth_down=2):
        """
        Create a proximity graph centered around the specified node.

        @type  center_node_id: pgraph.node
        @param center_node_id: Node to use as center of proximity graph
        @type  max_depth_up:   Integer
        @param max_depth_up:   (Optional, Def=2) Number of upward levels to include in proximity graph
        @type  max_depth_down: Integer
        @param max_depth_down: (Optional, Def=2) Number of downward levels to include in proximity graph

        @rtype:  pgraph.Graph
        @return: Proximity graph around specified node.
        """

        prox_graph = self.graph_down(center_node_id, max_depth_down)
        prox_graph.add_graph(self.graph_up(center_node_id, max_depth_up))

        return prox_graph

    def graph_sub(self, other_graph):
        """
        Remove the elements shared between the current graph and other graph from the current
        graph.

        @todo: Add support for clusters

        @type  other_graph: pgraph.Graph
        @param other_graph: Graph to diff/remove against
        """

        for other_node in list(other_graph.nodes.values()):
            self.del_node(other_node.id)

        for other_edge in list(other_graph.edges.values()):
            self.del_edge(None, other_edge.src, other_edge.dst)

        return self

    def graph_up(self, from_node_id, max_depth=-1):
        """
        Create a new graph, looking up, from the specified node id to the specified depth.

        @type  from_node_id: pgraph.node
        @param from_node_id: Node to use as start of up graph
        @type  max_depth:    Integer
        @param max_depth:    (Optional, Def=-1) Number of levels to include in up graph (-1 for infinite)

        @rtype:  pgraph.Graph
        @return: Up graph to the specified node.
        """

        up_graph = Graph()
        from_node = self.find_node("id", from_node_id)

        levels_to_process = []
        current_depth = 1

        levels_to_process.append([from_node])

        for level in levels_to_process:
            next_level = []

            # noinspection PyChainedComparisons
            if current_depth > max_depth and max_depth != -1:
                break

            for node in level:
                up_graph.add_node(copy.copy(node))

                for edge in self.edges_to(node.id):
                    to_add = self.find_node("id", edge.src)

                    if not up_graph.find_node("id", edge.src):
                        next_level.append(to_add)

                    up_graph.add_node(copy.copy(to_add))
                    up_graph.add_edge(copy.copy(edge))

            if next_level:
                levels_to_process.append(next_level)

            current_depth += 1

        return up_graph

    def render_graph_gml(self):
        """
        Render the GML graph description.

        :returns: GML graph description.
        :rtype: str
        """

        gml = 'Creator "pGRAPH - Pedram Amini <pedram.amini@gmail.com>"\n'
        gml += "directed 1\n"

        # open the graph tag.
        gml += "graph [\n"

        # add the nodes to the GML definition.
        for node in list(self.nodes.values()):
            gml += node.render_node_gml()

        # add the edges to the GML definition.
        for edge in list(self.edges.values()):
            gml += edge.render_edge_gml(self)

        # close the graph tag.
        gml += "\n]\n"

        """
        TODO: Complete cluster rendering
        # if clusters exist.
        if len(self.clusters):
            # open the rootcluster tag.
            gml += 'rootcluster [\n'

            # add the clusters to the GML definition.
            for cluster in self.clusters:
                gml += cluster.render()

            # add the clusterless nodes to the GML definition.
            for node in self.nodes:
                if not self.find_cluster_by_node("id", node.id):
                    gml += '    vertex "%d"\n' % node.id

            # close the rootcluster tag.
            gml += ']\n'
        """

        return gml

    def render_graph_graphviz(self):
        """
        Render the graphviz graph structure.

        Example to create a png:

        .. code-block::

            with open('somefile.png', 'wb') as file:
                file.write(session.render_graph_graphviz().create_png())

        :returns: Pydot object representing entire graph
        :rtype: pydot.Dot
        """
        dot_graph = pydot.Dot()

        for node in list(self.nodes.values()):
            dot_graph.add_node(node.render_node_graphviz())

        for edge in list(self.edges.values()):
            dot_graph.add_edge(edge.render_edge_graphviz())

        return dot_graph

    def render_graph_udraw(self):
        """
        Render the uDraw graph description.

        :returns: uDraw graph description.
        :rtype: str
        """

        udraw = "["

        # render each of the nodes in the graph.
        # the individual nodes will handle their own edge rendering.
        for node in list(self.nodes.values()):
            udraw += node.render_node_udraw(self)
            udraw += ","

        # trim the extraneous comment and close the graph.
        udraw = udraw[0:-1] + "\n]"

        return udraw

    def render_graph_udraw_update(self):
        """
        Render the uDraw graph update description.

        :returns: uDraw graph description.
        :rtype: str
        """

        udraw = "["

        for node in list(self.nodes.values()):
            udraw += node.render_node_udraw_update()
            udraw += ","

        for edge in list(self.edges.values()):
            udraw += edge.render_edge_udraw_update()
            udraw += ","

        # trim the extraneous comment and close the graph.
        udraw = udraw[0:-1] + "]"

        return udraw

    def update_node_id(self, current_id, new_id):
        """
        Simply updating the id attribute of a node will sever the edges to / from the given node. This routine will
        correctly update the edges as well.

        @type  current_id: long
        @param current_id: Current ID of node whose ID we want to update
        @type  new_id:     long
        @param new_id:     New ID to update to.
        """

        if current_id not in self.nodes:
            return

        # update the node.
        node = self.nodes[current_id]
        del self.nodes[current_id]
        node.id = new_id
        self.nodes[node.id] = node

        # update the edges.
        for edge in [edge for edge in list(self.edges.values()) if current_id in (edge.src, edge.dst)]:
            del self.edges[edge.id]

            if edge.src == current_id:
                edge.src = new_id
            if edge.dst == current_id:
                edge.dst = new_id

            edge.id = (edge.src << 32) + edge.dst

            self.edges[edge.id] = edge

    def sorted_nodes(self):
        """
        Return a list of the nodes within the graph, sorted by id.

        @rtype:  List
        @return: List of nodes, sorted by id.
        """

        node_keys = list(self.nodes)
        node_keys.sort()

        return [self.nodes[key] for key in node_keys]



import re

import flask
from flask import Flask, redirect, render_template

from .. import exception

MAX_LOG_LINE_LEN = 1500

app = Flask(__name__)
app.session = None


def commify(number):
    number = str(number)
    processing = 1
    regex = re.compile(r"^(-?\d+)(\d{3})")
    while processing:
        (number, processing) = regex.subn(r"\1,\2", number)
    return number


@app.route("/togglepause")
def pause():
    # Flip our state
    app.session.is_paused = not app.session.is_paused
    return redirect("/")


@app.route("/test-case/<int:crash_id>")
def test_case(crash_id):
    return render_template(
        "test-case.html",
        crashinfo=app.session.procmon_results.get(crash_id, None),
        test_case=app.session.test_case_data(crash_id),
    )


@app.route("/api/current-test-case")
def current_test_case_update():
    data = {"index": app.session.total_mutant_index, "log_data": _get_log_data(app.session.total_mutant_index)}
    return flask.jsonify(data)


@app.route("/api/test-case/<int:test_case_index>")
def api_test_case(test_case_index):
    data = {"index": test_case_index, "log_data": _get_log_data(test_case_id=test_case_index)}
    return flask.jsonify(data)


def _get_log_data(test_case_id):
    results = []
    try:
        case = app.session.test_case_data(test_case_id)
    except exception.BoofuzzNoSuchTestCase:
        return None
    if case is not None:
        results.append({"css_class": case.css_class, "log_line": case.html_log_line})
        for step in case.steps:
            line = step.html_log_line
            results.append({"css_class": step.css_class, "log_line": line})
    return results


@app.route("/api/current-run")
def index_update():
    data = {
        "session_info": {
            "is_paused": app.session.is_paused,
            "current_index": app.session.total_mutant_index,
            "num_mutations": app.session.total_num_mutations,
            "current_index_element": app.session.mutant_index if app.session is not None else None,
            "num_mutations_element": app.session.fuzz_node.get_num_mutations()
            if app.session.fuzz_node is not None
            else None,
            "current_element": app.session.fuzz_node.name if app.session.fuzz_node is not None else None,
            "current_test_case_name": app.session.current_test_case_name,
            "crashes": _crash_summary_info(),
            "runtime": app.session.runtime,
            "exec_speed": app.session.exec_speed,
        }
    }

    return flask.jsonify(data)


@app.route("/")
def index():
    crashes = _crash_summary_info()

    # which node (request) are we currently fuzzing.
    if app.session.fuzz_node is not None and app.session.fuzz_node.name:
        current_name = app.session.fuzz_node.name
    else:
        current_name = "[N/A]"

    # render sweet progress bars.
    if app.session.fuzz_node is not None:
        mutant_index = float(app.session.mutant_index)
        num_mutations = float(app.session.fuzz_node.get_num_mutations())

        try:
            progress_current = min(mutant_index / num_mutations, 1)
        except ZeroDivisionError:
            progress_current = 0
        num_bars = int(progress_current * 50)
        progress_current_bar = "[" + "=" * num_bars + "&nbsp;" * (50 - num_bars) + "]"
        progress_current = "%.3f%%" % (progress_current * 100)
    else:
        progress_current = 0
        progress_current_bar = ""
        mutant_index = 0
        num_mutations = 100  # TODO improve template instead of hard coding fake values

    total_mutant_index = float(app.session.total_mutant_index)
    total_num_mutations = app.session.total_num_mutations
    if total_num_mutations is None:
        progress_total = 0
    else:
        try:
            progress_total = min(total_mutant_index / total_num_mutations, 1)
        except ZeroDivisionError:
            progress_total = 0

    num_bars = int(progress_total * 50)
    progress_total_bar = "[" + "=" * num_bars + "&nbsp;" * (50 - num_bars) + "]"
    progress_total = "%.3f%%" % (progress_total * 100)

    state = {
        "session": app.session,
        "current_mutant_index": commify(int(mutant_index)),
        "current_name": current_name,
        "current_num_mutations": commify(int(num_mutations)),
        "progress_current": progress_current,
        "progress_current_bar": progress_current_bar,
        "progress_total": progress_total,
        "progress_total_bar": progress_total_bar,
        "total_mutant_index": commify(int(total_mutant_index)),
        "total_num_mutations": commify(int(total_num_mutations)) if total_num_mutations is not None else None,
    }

    return render_template("index.html", state=state, crashes=crashes)


def _crash_summary_info():
    crashes = []
    procmon_result_keys = list(app.session.monitor_results)
    procmon_result_keys.sort()
    for key in procmon_result_keys:
        val = app.session.monitor_results[key]
        status_bytes = "&nbsp;"

        if key in app.session.monitor_data:
            status_bytes = commify(app.session.netmon_results[key])

        crash = {"key": key, "reasons": val, "status_bytes": status_bytes}
        crashes.append(crash)
    return crashes



import os
import shlex
import subprocess
import time
from builtins import str

from boofuzz import utils
from boofuzz.monitors.base_monitor import BaseMonitor


def _split_command_if_str(command):
    """Splits a shell command string into a list of arguments.

    If any individual item is not a string, item is returned unchanged.

    Designed for use with subprocess.Popen.

    Args:
        command (Union[basestring, :obj:`list` of :obj:`basestring`]): List of commands. Each command
        should be a string or a list of strings.

    Returns:
        (:obj:`list` of :obj:`list`: of :obj:`str`): List of lists of command arguments.
    """
    if isinstance(command, str):
        return shlex.split(command, posix=(os.name == "posix"))

    else:
        return command


class ProcessMonitorLocal(BaseMonitor):
    def __init__(self, crash_filename, debugger_class, proc_name=None, pid_to_ignore=None, level=1, coredump_dir=None):
        """
        @type  crash_filename: str
        @param crash_filename: Name of file to (un)serialize crash bin to/from
        @type  proc_name:      str
        @param proc_name:      (Optional, def=None) Process name to search for and attach to
        @type  pid_to_ignore:  int
        @param pid_to_ignore:  (Optional, def=None) Ignore this PID when searching for the target process
        @type  level:          int
        @param level:          (Optional, def=1) Log output level, increase for more verbosity
        """

        self.crash_filename = os.path.abspath(crash_filename)
        self.debugger_class = debugger_class
        self.proc_name = proc_name
        self.ignore_pid = pid_to_ignore
        self.log_level = level
        self.capture_output = False

        self.stop_commands = []
        self.start_commands = []
        self.test_number = None
        self.debugger_thread = None
        self.crash_bin = utils.crash_binning.CrashBinning()

        self.last_synopsis = ""

        self.coredump_dir = coredump_dir

        if not os.access(os.path.dirname(self.crash_filename), os.X_OK):
            self.log("invalid path specified for crash bin: %s" % self.crash_filename)
            raise Exception

        self.log("Process Monitor PED-RPC server initialized:")
        # self.log("\t listening on:  %s:%s" % (host, port))
        self.log("\t crash file:    %s" % self.crash_filename)
        self.log("\t # records:     %d" % len(self.crash_bin.bins))
        self.log("\t proc name:     %s" % self.proc_name)
        self.log("\t log level:     %d" % self.log_level)
        self.log("awaiting requests...")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.debugger_thread is not None and self.debugger_thread.is_alive():
            self.debugger_thread.stop_target()

    # noinspection PyMethodMayBeStatic
    def alive(self):
        """
        Returns True. Useful for PED-RPC clients who want to see if the PED-RPC connection is still alive.
        """

        return True

    def get_crash_synopsis(self):
        """
        Return the last recorded crash synopsis.

        @rtype:  String
        @return: Synopsis of last recorded crash.
        """
        # Since crash synopsis is called only after a failure, check for failures again:
        self.debugger_thread.post_send()

        return self.last_synopsis

    def log(self, msg="", level=1):
        """
        If the supplied message falls under the current log level, print the specified message to screen.

        @type  msg: str
        @param msg: Message to log
        """

        if self.log_level >= level:
            print("[%s] %s" % (time.strftime("%I:%M.%S"), msg))

    def post_send(self, **kwargs):
        """
        This routine is called after the fuzzer transmits a test case and returns the status of the target.

        Returns:
            bool: True if the target is still active, False otherwise.
        """
        # note: kwargs for pedrpc client compatibility
        if self.debugger_thread is not None:
            return self.debugger_thread.post_send()
        else:
            raise Exception("post_send called before pre_send!")

    def pre_send(self, *args, **kwargs):
        """
        This routine is called before the fuzzer transmits a test case and ensure the debugger thread is operational.

        @type  test_number: Integer
        @param test_number: Test number to retrieve PCAP for.
        """
        if len(args) > 0:
            test_number = args[0]
        else:
            session = kwargs["session"]
            test_number = session.total_mutant_index
        # note: kwargs for pedrpc client compatibility
        self.log("pre_send(%d)" % test_number, 10)
        self.test_number = test_number

        if self.debugger_thread is None or not self.debugger_thread.is_alive():
            self.start_target()
            self.debugger_thread.pre_send()

    def start_target(self):
        """
        Start up the target process by issuing the commands in self.start_commands.

        @returns True if successful.
        """
        self.log("local Starting target...")
        self._stop_target_if_running()
        self.log("creating debugger thread", 5)
        self.debugger_thread = self.debugger_class(
            self.start_commands,
            self,
            proc_name=self.proc_name,
            ignore_pid=self.ignore_pid,
            log_level=self.log_level,
            coredump_dir=self.coredump_dir,
            capture_output=self.capture_output,
        )
        self.debugger_thread.daemon = True
        self.debugger_thread.start()
        self.debugger_thread.finished_starting.wait()
        self.log("giving debugger thread 2 seconds to settle in", 5)
        time.sleep(2)
        return True

    def stop_target(self):
        """
        Kill the current debugger thread and stop the target process by issuing the commands in self.stop_commands.
        """
        self.log("Stopping target...")

        if self._target_is_running():
            self._stop_target()
            self.log("target stopped")
        else:
            self.log("target already stopped")

    def _stop_target_if_running(self):
        """Stop target, if it is running. Return true if it was running; otherwise false."""
        if self._target_is_running():
            self.log("target still running; stopping first...")
            self._stop_target()
            self.log("target stopped")
            return True
        else:
            return False

    def _stop_target(self):
        # give the debugger thread a chance to exit.
        time.sleep(1)
        if len(self.stop_commands) < 1:
            self.debugger_thread.stop_target()
            while self.debugger_thread.is_alive():
                time.sleep(0.1)
        else:
            for command in self.stop_commands:
                if command == ["TERMINATE_PID"] or command == "TERMINATE_PID":
                    self.debugger_thread.stop_target()
                    while self.debugger_thread.is_alive():
                        time.sleep(0.1)
                else:
                    self.log("Executing stop command: '{0}'".format(command), 2)
                    subprocess.Popen(command)

    def _target_is_running(self):
        return self.debugger_thread is not None and self.debugger_thread.is_alive()

    def restart_target(self, **kwargs):
        """
        Stop and start the target process.

        @returns True if successful.
        """
        # note: kwargs for pedrpc client compatibility
        self.log("Restarting target...")
        self.stop_target()
        return self.start_target()

    def set_capture_output(self, capture_output):
        self.log("updating capture_output to '%s'" % capture_output)
        self.capture_output = capture_output

    def set_proc_name(self, new_proc_name):
        self.log("updating target process name to '%s'" % new_proc_name)
        self.proc_name = new_proc_name

    def set_start_commands(self, new_start_commands):
        self.log("updating start commands to: {0}".format(list(new_start_commands)))
        self.start_commands = list(map(_split_command_if_str, new_start_commands))

    def set_stop_commands(self, new_stop_commands):
        self.log("updating stop commands to: {0}".format(list(new_stop_commands)))
        self.stop_commands = new_stop_commands
        self.stop_commands = list(map(_split_command_if_str, new_stop_commands))

    def set_crash_filename(self, new_crash_filename):
        self.log("updating crash bin filename to '%s'" % new_crash_filename)
        self.crash_filename = new_crash_filename

    def set_options(self, *args, **kwargs):
        """
        Compatibility method to act like a pedrpc client.
        """
        # args will be ignored, kwargs will be translated

        for arg, value in kwargs.items():
            setattr(self, arg, value)

    def post_start_target(self, *args, **kwargs):
        """
        Compatibility method to act like a pedrpc client.
        """
        return



import abc


# abc.ABCMeta is the metaclass in both python 2 and 3
class ISerialLike(metaclass=abc.ABCMeta):
    """
    A serial-like interface, based on the pySerial module,
    the notable difference being that open() must always be called after the object is first created.

    Facilitates dependency injection in modules that use pySerial.

    .. versionchanged:: 0.2.0
        ISerialLike has been moved into the connections subpackage.
        The full path is now boofuzz.connections.iserial_like.ISerialLike
    """

    @abc.abstractmethod
    def close(self):
        """
        Close connection to the target.

        :return: None
        """
        raise NotImplementedError

    @abc.abstractmethod
    def open(self):
        """
        Opens connection to the target. Make sure to call close!

        :return: None
        """
        raise NotImplementedError

    @abc.abstractmethod
    def recv(self, max_bytes):
        """
        Receive up to max_bytes data from the target.

        :param max_bytes: Maximum number of bytes to receive.
        :type max_bytes: int

        :return: Received data.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def send(self, data):
        """
        Send data to the target. Only valid after calling open!

        :param data: Data to send.

        :return: Number of bytes actually sent.
        """
        raise NotImplementedError



import errno
import importlib.metadata
import os
import re
import signal
import socket
import struct
import time
import warnings
import zlib
from builtins import int
from functools import reduce

from colorama import Back, Fore, Style

from boofuzz.connections import ip_constants, udp_socket_connection
from boofuzz.exception import BoofuzzError

# Curses color pairs

COLOR_PAIR_WHITE = 1
COLOR_PAIR_CYAN = 2
COLOR_PAIR_RED = 3
COLOR_PAIR_YELLOW = 4
COLOR_PAIR_GREEN = 5
COLOR_PAIR_MAGENTA = 6
COLOR_PAIR_BLACK = 7

test_step_info = {
    "test_case": {
        "indent": 0,
        "title": "Test Case",
        "html": "Test Case: {msg}",
        "terminal": Fore.YELLOW + Style.BRIGHT + "Test Case: {msg}" + Style.RESET_ALL,
        "css_class": "log-case",
        "curses": COLOR_PAIR_YELLOW,
    },
    "step": {
        "indent": 1,
        "title": "Test Step",
        "html": " Test Step: {msg}",
        "terminal": Fore.MAGENTA + Style.BRIGHT + "Test Step: {msg}" + Style.RESET_ALL,
        "css_class": "log-step",
        "curses": COLOR_PAIR_MAGENTA,
    },
    "info": {
        "indent": 2,
        "title": "Info",
        "html": "Info: {msg}",
        "terminal": "Info: {msg}",
        "css_class": "log-info",
        "curses": COLOR_PAIR_WHITE,
    },
    "error": {
        "indent": 2,
        "title": "Error",
        "html": "Error!!!! {msg}",
        "terminal": Back.RED + Style.BRIGHT + "Error!!!! {msg}" + Style.RESET_ALL,
        "css_class": "log-error",
        "curses": COLOR_PAIR_RED,
    },
    "send": {
        "indent": 2,
        "title": "Transmitted",
        "html": "Transmitted {n} bytes{note}: {msg}",
        "terminal": Fore.CYAN + "Transmitted {n} bytes{note}: {msg}" + Style.RESET_ALL,
        "css_class": "log-send",
        "curses": COLOR_PAIR_CYAN,
    },
    "receive": {
        "indent": 2,
        "title": "Received",
        "html": "Received{note}: {msg}",
        "terminal": Fore.CYAN + "Received{note}: {msg}" + Style.RESET_ALL,
        "css_class": "log-receive",
        "curses": COLOR_PAIR_CYAN,
    },
    "check": {
        "indent": 2,
        "title": "Check",
        "html": "Check: {msg}",
        "terminal": "Check: {msg}",
        "css_class": "log-check",
        "curses": COLOR_PAIR_WHITE,
    },
    "fail": {
        "indent": 3,
        "title": "Check Failed",
        "html": "Check Failed: {msg}",
        "terminal": Fore.RED + Style.BRIGHT + "Check Failed: {msg}" + Style.RESET_ALL,
        "css_class": "log-fail",
        "curses": COLOR_PAIR_RED,
    },
    "pass": {
        "indent": 3,
        "title": "Check OK",
        "html": "Check OK: {msg}",
        "terminal": Fore.GREEN + Style.BRIGHT + "Check OK: {msg}" + Style.RESET_ALL,
        "css_class": "log-pass",
        "curses": COLOR_PAIR_GREEN,
    },
}


def ip_str_to_bytes(ip):
    """Convert an IP string to a four-byte bytes.

    :param ip: IP address string, e.g. '127.0.0.1'

    :return 4-byte representation of ip, e.g. b'\x7F\x00\x00\x01'
    :rtype bytes

    :raises ValueError if ip is not a legal IP address.
    """
    try:
        return socket.inet_aton(ip)
    except socket.error:
        raise ValueError("Illegal IP address passed to socket.inet_aton: {0}".format(ip))


def get_max_udp_size():
    """
    Crazy CTypes magic to do a getsockopt() which determines the max UDP payload size in a platform-agnostic way.

    .. deprecated:: 0.2.0
        Use :meth:`UDPSocketConnection.max_payload() <boofuzz.connections.UDPSocketConnection.max_payload>` instead.

    Returns:
        int: The maximum length of a UDP packet the current platform supports
    """
    warnings.warn(
        "get_max_udp_size() is deprecated and will be removed in a future version of boofuzz. "
        "Use UDPSocketConnection.max_payload() instead.",
        FutureWarning,
    )

    return udp_socket_connection.UDPSocketConnection.max_payload()


def calculate_four_byte_padding(string, character="\x00"):
    return character * ((4 - (len(string) & 3)) & 3)


def crc16(string, value=0):
    """CRC-16 poly: p(x) = x**16 + x**15 + x**2 + 1

    @param string: Data over which to calculate crc.
    @param value: Initial CRC value.
    """
    crc16_table = []
    for byte in range(256):
        crc = 0

        for _ in range(8):
            if (byte ^ crc) & 1:
                crc = (crc >> 1) ^ 0xA001  # polly
            else:
                crc >>= 1

            byte >>= 1

        crc16_table.append(crc)

    for ch in string:
        value = crc16_table[ord(ch) ^ (value & 0xFF)] ^ (value >> 8)

    return value


def crc32(string):
    return zlib.crc32(string) & 0xFFFFFFFF


def uuid_bin_to_str(uuid):
    """Convert a binary UUID to human readable string.

    @param uuid: bytes representing UUID.
    """
    (block1, block2, block3) = struct.unpack("<LHH", uuid[:8])
    (block4, block5, block6) = struct.unpack(">HHL", uuid[8:16])

    return "%08x-%04x-%04x-%04x-%04x%08x" % (block1, block2, block3, block4, block5, block6)


def uuid_str_to_bin(uuid):
    """
    Converts a UUID string to binary form.

    Expected string input format is same as uuid_bin_to_str()'s output format.

    Ripped from Core Impacket.

    :param uuid: UUID string to convert to bytes.
    :type uuid: str
    :return: UUID as bytes.
    :rtype: bytes
    """
    uuid_re = r"([\dA-Fa-f]{8})-([\dA-Fa-f]{4})-([\dA-Fa-f]{4})-([\dA-Fa-f]{4})-([\dA-Fa-f]{4})([\dA-Fa-f]{8})"

    matches = re.match(uuid_re, uuid)

    # pytype: disable=attribute-error
    (uuid1, uuid2, uuid3, uuid4, uuid5, uuid6) = map(lambda x: int(x, 16), matches.groups())
    # pytype: enable=attribute-error

    uuid = struct.pack("<LHH", uuid1, uuid2, uuid3)
    uuid += struct.pack(">HHL", uuid4, uuid5, uuid6)

    return uuid


def _ones_complement_sum_carry_16(a, b):
    """Compute ones complement sum and carry at 16 bits.

    :type a: int
    :type b: int

    :return: Sum of a and b, ones complement, carry at 16 bits.
    """
    pre_sum = a + b
    return (pre_sum & 0xFFFF) + (pre_sum >> 16)


def _collate_bytes(msb, lsb):
    """
    Helper function for our helper functions.
    Collates msb and lsb into one 16-bit value.

    :type msb: byte
    :param msb: Single byte (most significant).

    :type lsb: byte
    :param lsb: Single byte (least significant).

    :return: msb and lsb all together in one 16 bit value.
    """
    return (msb << 8) + lsb


def ipv4_checksum(msg):
    """
    Return IPv4 checksum of msg.
    :param msg: Message to compute checksum over.
    :type msg: bytes

    :return: IPv4 checksum of msg.
    :rtype: int
    """
    # Pad with 0 byte if needed
    if len(msg) % 2 == 1:
        msg += b"\x00"

    msg_words = map(_collate_bytes, msg[0::2], msg[1::2])
    total = reduce(_ones_complement_sum_carry_16, msg_words, 0)
    return ~total & 0xFFFF


def _udp_checksum_pseudo_header(src_addr, dst_addr, msg_len):
    """Return pseudo-header for UDP checksum.

    :type src_addr: bytes
    :param src_addr: Source IP address -- 4 bytes.

    :type dst_addr: bytes
    :param dst_addr: Destination IP address -- 4 bytes.

    :param msg_len: Length of UDP message (not including IPv4 header).
    :type msg_len: int

    :return: UDP pseudo-header
    :rtype: bytes
    """
    return src_addr + dst_addr + b"\x00" + bytes([ip_constants.IPV4_PROTOCOL_UDP]) + struct.pack(">H", msg_len)


def udp_checksum(msg, src_addr, dst_addr):
    """Return UDP checksum of msg.

    Recall that the UDP checksum involves creating a sort of pseudo IP header.
    This header requires the source and destination IP addresses, which this
    function takes as parameters.

    If msg is too big, the checksum is undefined, and this method will
    truncate it for the sake of checksum calculation. Note that this means the
    checksum will be invalid. This loosey goosey error checking is done to
    support fuzz tests which at times generate huge, invalid packets.


    :param msg: Message to compute checksum over.
    :type msg: bytes

    :type src_addr: bytes
    :param src_addr: Source IP address -- 4 bytes.
    :type dst_addr: bytes
    :param dst_addr: Destination IP address -- 4 bytes.

    :return: UDP checksum of msg.
    :rtype: int
    """
    # If the packet is too big, the checksum is undefined since len(msg)
    # won't fit into two bytes. So we just pick our best definition.
    # "Truncate" the message as it appears in the checksum.
    msg = msg[0 : ip_constants.UDP_MAX_LENGTH_THEORETICAL]

    return ipv4_checksum(_udp_checksum_pseudo_header(src_addr, dst_addr, len(msg)) + msg)


def hex_str(s):
    """
    Returns a hex-formatted string based on s.

    Args:
        s (bytes): Some string.

    Returns:
        str: Hex-formatted string representing s.
    """
    return " ".join("{:02x}".format(b) for b in bytearray(s))


def pause_for_signal():
    """
    Pauses the current thread in a way that can still receive signals like SIGINT from Ctrl+C.

    Implementation notes:
     - Linux uses signal.pause()
     - Windows uses a loop that sleeps for 1 ms at a time, allowing signals
       to interrupt the thread fairly quickly.

    :return: None
    :rtype: None
    """
    try:
        while True:
            signal.pause()
    except AttributeError:
        # signal.pause() is missing for Windows; wait 1ms and loop instead
        while True:
            time.sleep(0.001)


def get_time_stamp():
    t = time.time()
    s = time.strftime("[%Y-%m-%d %H:%M:%S", time.localtime(t))
    s += ",%03d]" % (t * 1000 % 1000)
    return s


def _indent_all_lines(lines, amount, ch=" "):
    padding = amount * ch
    return padding + ("\n" + padding).join(lines.split("\n"))


def _indent_after_first_line(lines, amount, ch=" "):
    padding = amount * ch
    return ("\n" + padding).join(lines.split("\n"))


def format_log_msg(
    msg_type, description=None, data=None, indent_size=2, timestamp=None, truncated=False, format_type="terminal"
):
    curses_mode = False
    if data is None:
        data = b""
    if timestamp is None:
        timestamp = get_time_stamp()
    if format_type == "curses":
        curses_mode = True
        format_type = "html"

    if description is not None and description != "":
        msg = description
    elif data is not None and len(data) > 0:
        msg = hex_to_hexstr(input_bytes=data)
    else:
        msg = ""

    msg = test_step_info[msg_type][format_type].format(  # pytype: disable=attribute-error
        msg=msg, n=len(data), note="" if not truncated else " (data truncated for database storage)"
    )
    msg = _indent_all_lines(msg, (test_step_info[msg_type]["indent"]) * indent_size)
    msg = timestamp + " " + _indent_after_first_line(msg, len(timestamp) + 1)

    if curses_mode:
        return [msg, test_step_info[msg_type]["curses"]]

    return msg


def format_msg(msg, indent_level, indent_size, timestamp=None):
    msg = _indent_all_lines(msg, indent_level * indent_size)
    if timestamp is None:
        timestamp = get_time_stamp()
    return timestamp + " " + _indent_after_first_line(msg, len(timestamp) + 1)


def hex_to_hexstr(input_bytes):
    """
    Render input_bytes as ASCII-encoded hex bytes, followed by a best effort
    utf-8 rendering.

    Args:
        input_bytes (bytes): Arbitrary bytes

    Returns:
        str: Printable string
    """
    return hex_str(input_bytes) + " " + repr(input_bytes)


def mkdir_safe(directory_name, file_included=False):
    """Creates directory_name and subdirectories. If file_included is true, removes final element of the path"""
    if file_included:
        fullpath = os.path.abspath(directory_name)
        directory_name = os.path.dirname(fullpath)
    try:
        os.makedirs(directory_name)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def path_exists(path):
    """To avoid polluting files with import os"""
    return os.path.exists(path)


def get_boofuzz_version():
    """
    Gets the currently installed boofuzz version

    :rtype: str
    :return: Boofuzz version as string
    """
    return "v" + importlib.metadata.version("boofuzz")


def str_to_bytes(value, encoding="utf-8", errors="replace"):
    if isinstance(value, bytes):
        return value
    return value.encode(encoding, errors)


def parse_target(target_name):
    try:
        host, ip = target_name.split(":")
        return host, int(ip)
    except ValueError:
        raise ValueError("Target format is HOST:PORT")


def parse_test_case_name(test_case):
    """Parse a test case name into a message path and a list of mutation names.

    Example:
        Input: "message1:[message1.first_byte:2, message1.second_byte:1, message1.third_byte:2]"
        Output: ["message1"], ["message1.first_byte:2", "message1.second_byte:1", "message1.third_byte:2"]

    Returns:
        A message path (list of message names) and a list of mutation names.
    """
    components = test_case.split(":", 1)
    message_path = components[0]
    path = re.split("->", message_path)
    if len(components) < 2:
        return path, []
    else:
        mutations = components[1]
        match = re.match(r"\[(.*)\]", mutations)
        if match is None:
            raise BoofuzzError("could not parse test case name: {0}".format(test_case))
        mutations = match.group(1)
        mutations = re.split(r",\s*", mutations)
        return path, mutations



import attr


@attr.s
class BoofuzzFailure(Exception):
    """Raises a failure in the current test case.

    This is most important when failures in the target may not be noticed until the next test case. In such a situation,
    it may be necessary to abort the test case before a fuzz message is even sent, as the fuzz message may be poorly
    defined outside the context of a valid partial protocol exchange.
    """

    message = attr.ib(type=str, default="")
    pass


class BoofuzzError(Exception):
    pass


class BoofuzzRestartFailedError(BoofuzzError):
    pass


class BoofuzzTargetConnectionFailedError(BoofuzzError):
    pass


class BoofuzzOutOfAvailableSockets(BoofuzzError):
    pass


class BoofuzzTargetConnectionReset(BoofuzzError):
    pass


@attr.s
class BoofuzzTargetConnectionAborted(BoofuzzError):
    """
    Raised on `errno.ECONNABORTED`.
    """

    socket_errno = attr.ib()
    socket_errmsg = attr.ib()


class BoofuzzNoSuchTestCase(BoofuzzError):
    pass


class BoofuzzRpcError(BoofuzzError):
    pass


class SullyRuntimeError(Exception):
    pass


class SizerNotUtilizedError(Exception):
    pass


class MustImplementException(Exception):
    pass


class BoofuzzSSLError(BoofuzzError):
    pass


class BoofuzzNameResolutionError(BoofuzzError):
    pass



import random
import struct

from ..fuzzable import Fuzzable


class Float(Fuzzable):
    """Primitive that generates random float values within a specific range and with a fixed format.

    :type name: str, optional
    :param name: Name, for referencing later.
    :type default_value: float
    :param default_value: Value used when the element is not being fuzzed.
    :type s_format: str, optional
    :param s_format: Format of the float value on encoding, defaults to .1f
    :type f_min: float, optional
    :param f_min: Minimal float value that can be generated while fuzzing, defaults to 0.0
    :type f_max: float, optional
    :param f_max: Maximal float value that can be generated while fuzzing, defaults to 100.0
    :type max_mutations: int, optional
    :param max_mutations: Total number of mutations for this individual primitive, defaults to 1000
    :type seed: int or str or bytes or bytearray
    :param seed: Set random.seed() with the given seed for reproducible results
    :type encode_as_ieee_754: bool, optional
    :param encode_as_ieee_754: Encode the float value as IEEE 754 floating point
    :type endian: str, optional
    :param endian: Change the endianness of IEEE 754 float point representation, defaults to big endian
    """

    def __init__(
        self,
        name=None,
        default_value=0.0,
        s_format=".1f",
        f_min=0.0,
        f_max=100.0,
        max_mutations=1000,
        seed=None,
        encode_as_ieee_754=False,
        endian="big",
        *args,
        **kwargs
    ):
        super(Float, self).__init__(name=name, default_value=str(default_value), *args, **kwargs)

        self.s_format = s_format
        self.f_min = f_min
        self.f_max = f_max
        self.max_mutations = max_mutations
        self.seed = seed
        self.encode_as_ieee_754 = encode_as_ieee_754
        self.endian = endian

    def mutations(self, default_value):
        last_val = None
        if self.seed is not None:
            random.seed(self.seed)

        for i in range(self.max_mutations):
            if i == 0:
                current_val = default_value
            else:
                current_val = random.uniform(self.f_min, self.f_max)

            str_format = "%" + self.s_format
            current_val = str_format % float(current_val)

            if last_val == current_val:
                continue
            last_val = current_val
            yield current_val

    def encode(self, value, mutation_context=None):
        if self.encode_as_ieee_754:
            value = float(value)
            return self.__convert_to_iee_754(value)

        return value.encode()

    def __convert_to_iee_754(self, value):
        if self.endian == "big":
            iee_value = struct.pack(">f", value)
        elif self.endian == "little":
            iee_value = struct.pack("<f", value)
        else:
            error_msg = "Invalid endian argument '%s'. Use 'big' or 'little'." % self.endian
            raise ValueError(error_msg)

        return iee_value

    def num_mutations(self, default_value):
        return self.max_mutations



#!/usr/bin/env python3
# Designed for use with boofuzz v0.2.0
#
# Minimal example which act as a TLS server
# It will listen on localhost and fuzz the connecting TLS client

import ssl

from boofuzz import *

# In order to act as a SSL/TLS server, boofuzz requires a SSL/TLS
# certificate. You can create a self-signed one with
#    openssl req -x509 -newkey rsa -keyout key.pem -out cert.pem -days 365 -nodes
ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
ctx.load_cert_chain(certfile="cert.pem", keyfile="key.pem")

session = Session(
    target=Target(connection=SSLSocketConnection(host="127.0.0.1", port=443, sslcontext=ctx, server=True))
)
s_initialize("A")
s_string("A")
session.connect(session.root, s_get("A"))
session.fuzz()



#!/usr/bin/env python3
# Designed for use with boofuzz v0.2.0

from boofuzz import *


def main():
    port = 69
    host = "127.0.0.1"

    session = Session(
        target=Target(
            connection=UDPSocketConnection(host, port),
        ),
    )

    s_initialize("RRQ")
    s_static("\x00\x01")
    s_string("filename", name="Filename")
    s_static("\x00")
    s_string("netascii", name="Mode")
    s_static("\x00")

    s_initialize("WRQ")
    s_static("\x00\x02")
    s_string("filename", name="Filename")
    s_static("\x00")
    s_string("netascii", name="Mode")
    s_static("\x00")

    s_initialize("TRQ")
    s_static("\x00\x02")
    s_string("filename", name="Filename")
    s_static("\x00")
    s_static("mail")
    s_static("\x00")

    session.connect(s_get("RRQ"))
    session.connect(s_get("WRQ"))
    session.connect(s_get("TRQ"))

    session.fuzz()


if __name__ == "__main__":
    main()



import functools
import operator
from collections.abc import ByteString

from funcy import compose

from ..fuzzable import Fuzzable


class Bytes(Fuzzable):
    """Primitive that fuzzes a binary byte string with arbitrary length.

    :type name: str, optional
    :param name: Name, for referencing later. Names should always be provided, but if not, a default name will be given,
        defaults to None
    :type default_value: bytes, optional
    :param default_value: Value used when the element is not being fuzzed - should typically represent a valid value,
        defaults to b""
    :type size: int, optional
    :param size: Static size of this field, leave None for dynamic, defaults to None
    :type padding: chr, optional
    :param padding: Value to use as padding to fill static field size, defaults to b"\\x00"
    :type max_len: int, optional
    :param max_len: Maximum string length, defaults to None
    :type fuzzable: bool, optional
    :param fuzzable: Enable/disable fuzzing of this primitive, defaults to true
    """

    # This binary strings will always included as testcases.
    _fuzz_library = [
        b"",
        b"\x00",
        b"\xFF",
        b"A" * 10,
        b"A" * 100,
        b"A" * 1000,
        b"A" * 5000,
        b"A" * 10000,
        b"A" * 100000,
    ]

    # from https://en.wikipedia.org/wiki/Magic_number_(programming)#Magic_debug_values
    _magic_debug_values = [
        b"\x00\x00\x81#",
        b"\x00\xfa\xca\xde",
        b"\x1b\xad\xb0\x02",
        b"\x8b\xad\xf0\r",
        b"\xa5\xa5\xa5\xa5",
        b"\xa5",
        b"\xab\xab\xab\xab",
        b"\xab\xad\xba\xbe",
        b"\xab\xba\xba\xbe",
        b"\xab\xad\xca\xfe",
        b"\xb1k\x00\xb5",
        b"\xba\xad\xf0\r",
        b"\xba\xaa\xaa\xad",
        b'\xba\xd2""',
        b"\xba\xdb\xad\xba\xdb\xad",
        b"\xba\xdc\x0f\xfe\xe0\xdd\xf0\r",
        b"\xba\xdd\xca\xfe",
        b"\xbb\xad\xbe\xef",
        b"\xbe\xef\xca\xce",
        b"\xc0\x00\x10\xff",
        b"\xca\xfe\xba\xbe",
        b"\xca\xfe\xd0\r",
        b"\xca\xfe\xfe\xed",
        b"\xcc\xcc\xcc\xcc",
        b"\xcd\xcd\xcd\xcd",
        b"\r\x15\xea^",
        b"\xdd\xdd\xdd\xdd",
        b"\xde\xad\x10\xcc",
        b"\xde\xad\xba\xbe",
        b"\xde\xad\xbe\xef",
        b"\xde\xad\xca\xfe",
        b"\xde\xad\xc0\xde",
        b"\xde\xad\xfa\x11",
        b"\xde\xad\xf0\r",
        b"\xde\xfe\xc8\xed",
        b"\xde\xad\xde\xad",
        b"\xeb\xeb\xeb\xeb",
        b"\xfa\xde\xde\xad",
        b"\xfd\xfd\xfd\xfd",
        b"\xfe\xe1\xde\xad",
        b"\xfe\xed\xfa\xce",
        b"\xfe\xee\xfe\xee",
    ]

    # This is a list of "interesting" 1,2 and 4 byte binary strings.
    # The lists are used to replace each block of 1, 2 or 4 byte in the original
    # value with each of those "interesting" values.
    _fuzz_strings_1byte = [b"\x00", b"\x01", b"\x7F", b"\x80", b"\xFF"] + [
        i for i in _magic_debug_values if len(i) == 1
    ]

    _fuzz_strings_2byte = [
        b"\x00\x00",
        b"\x01\x00",
        b"\x00\x01",
        b"\x7F\xFF",
        b"\xFF\x7F",
        b"\xFE\xFF",
        b"\xFF\xFE",
        b"\xFF\xFF",
    ] + [i for i in _magic_debug_values if len(i) == 2]

    _fuzz_strings_4byte = [
        b"\x00\x00\x00\x00",
        b"\x00\x00\x00\x01",
        b"\x01\x00\x00\x00",
        b"\x7F\xFF\xFF\xFF",
        b"\xFF\xFF\xFF\x7F",
        b"\xFE\xFF\xFF\xFF",
        b"\xFF\xFF\xFF\xFE",
        b"\xFF\xFF\xFF\xFF",
    ] + [i for i in _magic_debug_values if len(i) == 4]

    _mutators_of_default_value = [
        functools.partial(operator.mul, 2),
        functools.partial(operator.mul, 10),
        functools.partial(operator.mul, 100),
    ]

    def __init__(
        self,
        name: str = None,
        default_value: bytes = b"",
        size: int = None,
        padding: bytes = b"\x00",
        max_len: int = None,
        *args,
        **kwargs
    ):
        if not isinstance(default_value, ByteString):
            raise TypeError("default_value of Bytes must be of ByteString type")
        if not isinstance(padding, ByteString):
            raise TypeError("padding of Bytes must be of ByteString type")

        super(Bytes, self).__init__(name=name, default_value=default_value, *args, **kwargs)

        self.size = size
        self.max_len = max_len
        if self.size is not None:
            self.max_len = self.size
        self.padding = padding

    def mutations(self, default_value):
        for fuzz_value in self._iterate_fuzz_cases(default_value):
            if callable(fuzz_value):
                yield compose(self._adjust_mutation_for_size, fuzz_value)
            else:
                yield self._adjust_mutation_for_size(fuzz_value=fuzz_value)

    def _adjust_mutation_for_size(self, fuzz_value):
        if self.size is not None:
            if len(fuzz_value) > self.size:
                return fuzz_value[: self.max_len]
            else:
                return fuzz_value + self.padding * (self.size - len(fuzz_value))
        elif self.max_len is not None and len(fuzz_value) > self.max_len:
            return fuzz_value[: self.max_len]
        else:
            return fuzz_value

    def _iterate_fuzz_cases(self, default_value):
        for fuzz_value in self._fuzz_library:
            yield fuzz_value
        for fuzz_value in self._mutators_of_default_value:
            yield fuzz_value
        for fuzz_value in self._magic_debug_values:
            yield fuzz_value
        for i in range(0, len(default_value)):
            for fuzz_bytes in self._fuzz_strings_1byte:

                def f(value):
                    if i < len(value):
                        return value[:i] + fuzz_bytes + value[i + 1 :]
                    else:
                        return value

                yield f
        for i in range(0, len(default_value) - 1):
            for fuzz_bytes in self._fuzz_strings_2byte:

                def f(value):
                    if i < len(value) - 1:
                        return value[:i] + fuzz_bytes + value[i + 2 :]
                    else:
                        return value

                yield f

        for i in range(0, len(default_value) - 3):
            for fuzz_bytes in self._fuzz_strings_4byte:

                def f(value):
                    if i < len(value) - 3:
                        return value[:i] + fuzz_bytes + value[i + 4 :]
                    else:
                        return value

                yield f

    def num_mutations(self, default_value):
        """
        Calculate and return the total number of mutations for this individual primitive.

        @rtype:  int
        @return: Number of mutated forms this primitive can take
        :param default_value:
        """
        return sum(
            (
                len(self._fuzz_library),
                len(self._mutators_of_default_value),
                len(self._magic_debug_values),
                len(self._fuzz_strings_1byte) * max(0, len(default_value) - 0),
                len(self._fuzz_strings_2byte) * max(0, len(default_value) - 1),
                len(self._fuzz_strings_4byte) * max(0, len(default_value) - 3),
            )
        )

    def encode(self, value, mutation_context):
        if value is None:
            value = b""
        return value
